# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
import numpy as np
from typing import TYPE_CHECKING, List, Union, cast
from PySide6.QtCore import Qt, QPoint, QRectF, QPointF
from PySide6.QtGui import QIcon, QPixmap, QPainter, QPen, QBrush, QColor, QCursor, QTransform
from PySide6.QtWidgets import QMenu, QGraphicsItem, QGraphicsRectItem, QGraphicsSceneMouseEvent

from VeraGrid.Gui.gui_functions import add_menu_entry
from VeraGrid.Gui.Diagrams.SchematicWidget.terminal_item import RoundTerminalItem
from VeraGrid.Gui.Diagrams.generic_graphics import GenericDiagramWidget, ACTIVE, DEACTIVATED
from VeraGridEngine.Devices.Branches.vsc import VSC
from VeraGridEngine.Devices.Substation.bus import Bus
from VeraGridEngine.enumerations import DeviceType, ConverterControlType, \
    TerminalType  # Assuming VSC controls might be relevant later

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget
    from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.line_graphics_template import \
        LineGraphicTemplateItem  # For type hints if needed later
    from VeraGrid.Gui.Diagrams.SchematicWidget.terminal_item import RoundTerminalItem, BarTerminalItem


class VscGraphicItem3Term(GenericDiagramWidget, QGraphicsRectItem):
    """
    Graphics item for the 3-Terminal VSC converter
    """

    def __init__(self,
                 editor: SchematicWidget,
                 api_object: VSC,
                 pos: QPoint = None,
                 parent=None,
                 draw_labels: bool = True):
        """
        Constructor for the 3-Terminal VSC Graphic Item.

        :param editor: SchematicWidget instance
        :param api_object: VSC API object
        :param pos: Initial position
        :param parent: Parent item
        :param draw_labels: Whether to draw labels
        """
        GenericDiagramWidget.__init__(self, parent=parent, api_object=api_object, editor=editor,
                                      draw_labels=draw_labels)
        QGraphicsRectItem.__init__(self, parent=parent)

        # Make it a square
        self.w = 60  # Width of the VSC symbol
        self.h = self.w  # Height of the VSC symbol
        self.setRect(0.0, 0.0, self.w, self.h)

        self.draw_labels = draw_labels

        # Color and style based on active state
        self.pen_width = 2
        if self.api_object.active:
            self.color = ACTIVE['color']
            self.style = ACTIVE['style']
        else:
            self.color = DEACTIVATED['color']

        # Setup pen, brush, flags and cursor
        self.setPen(QPen(self.color, self.pen_width, self.style))
        self.setBrush(Qt.BrushStyle.NoBrush)  # Set transparent background
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
        self.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))

        if pos is not None:
            self.setPos(pos)
        else:
            self.setPos(QPoint(0, 0))

        # --- Create Terminals ---
        # Terminal positions (relative to the item's top-left corner 0,0)
        # AC terminal on the right middle
        t_ac_pos = QPointF(self.w, self.h / 2)
        # DC+ terminal on the left top
        t_dc_p_pos = QPointF(-10, self.h * 0.2)
        # DC- terminal on the left bottom
        t_dc_n_pos = QPointF(-10, self.h * 0.8)

        # self.terminals: List[RoundTerminalItem] = list()
        self.terminal_ac: RoundTerminalItem | None = None
        self.terminal_dc_p: RoundTerminalItem | None = None
        self.terminal_dc_n: RoundTerminalItem | None = None
        self.conn_line_ac: LineGraphicTemplateItem | None = None
        self.conn_line_dc_p: LineGraphicTemplateItem | None = None
        self.conn_line_dc_n: LineGraphicTemplateItem | None = None

        # AC Terminal (Index 0)
        self.terminal_ac = RoundTerminalItem("ac", parent=self, editor=self.editor, terminal_type=TerminalType.AC)
        self.terminal_ac.setPos(t_ac_pos)
        self.terminal_ac.setRotation(0)  # Points right
        self.terminal_ac.setPen(QPen(self.color, self.pen_width, self.style))

        # DC+ Terminal (Index 1)
        self.terminal_dc_p = RoundTerminalItem("dc_p", parent=self, editor=self.editor, terminal_type=TerminalType.DC_P)
        self.terminal_dc_p.setPos(t_dc_p_pos)
        self.terminal_dc_p.setRotation(0)  # Points left
        self.terminal_dc_p.setPen(QPen(self.color, self.pen_width, self.style))

        # DC- Terminal (Index 2)
        self.terminal_dc_n = RoundTerminalItem("dc_n", parent=self, editor=self.editor, terminal_type=TerminalType.DC_N)
        self.terminal_dc_n.setPos(t_dc_n_pos)
        self.terminal_dc_n.setRotation(0)  # Points left
        self.terminal_dc_n.setPen(QPen(self.color, self.pen_width, self.style))

        self.set_terminal_tooltips()

        # self.setRotation(180)

    @property
    def api_object(self) -> VSC:
        return self._api_object

    @property
    def editor(self) -> SchematicWidget:
        return self._editor

    def paint(self, painter: QPainter, option, widget=None) -> None:
        """Paint the VSC symbol."""
        # Draw the main rectangle (square)
        painter.setPen(QPen(self.color, self.pen_width, self.style))
        painter.setBrush(Qt.BrushStyle.NoBrush)  # Ensure transparent background
        painter.drawRect(self.rect())

        # Draw a diagonal line from top-right to bottom-left
        pen = QPen(self.color, self.pen_width)
        painter.setPen(pen)
        painter.drawLine(QPoint(self.w, 0), QPoint(0, self.h))

        # Draw AC/DC symbols inside the box near terminals
        text_rect_size = 15
        painter.drawText(
            QRectF(self.w * 0.8 - text_rect_size / 2, self.h / 2 - text_rect_size / 2 * 1.3, text_rect_size,
                   text_rect_size), Qt.AlignmentFlag.AlignCenter, "~")
        painter.drawText(QRectF(self.w * 0.15 - text_rect_size / 2, self.h * 0.35 - text_rect_size / 2, text_rect_size,
                                text_rect_size), Qt.AlignmentFlag.AlignCenter, "+")
        painter.drawText(QRectF(self.w * 0.15 - text_rect_size / 2, self.h * 0.65 - text_rect_size / 2, text_rect_size,
                                text_rect_size), Qt.AlignmentFlag.AlignCenter, "-")

    def set_terminal_tooltips(self):
        """Set tooltips for the terminals."""
        if self.api_object:
            self.terminal_ac.setToolTip(
                f"AC Terminal ({self.api_object.bus_to.name if self.api_object.bus_to else 'Unconnected'})")
            self.terminal_dc_p.setToolTip(
                f"DC+ Terminal ({self.api_object.bus_from.name if self.api_object.bus_from else 'Unconnected'})")
            self.terminal_dc_n.setToolTip(
                f"DC- Terminal ({self.api_object.bus_dc_n.name if self.api_object.bus_dc_n else 'Unconnected'})")

    def update_conn(self):
        """Update the connection lines attached to this item."""
        if self.conn_line_ac is not None:
            self.conn_line_ac.update_ports()
        if self.conn_line_dc_p is not None:
            self.conn_line_dc_p.update_ports()
        if self.conn_line_dc_n is not None:
            self.conn_line_dc_n.update_ports()

    def get_associated_widgets(self) -> List[LineGraphicTemplateItem]:
        """Return the graphical line items connected to this VSC."""
        return [self.conn_line_ac, self.conn_line_dc_p, self.conn_line_dc_n]

    def get_extra_graphics(self):
        """Return terminals associated with this widget."""
        return [self.terminal_ac, self.terminal_dc_p, self.terminal_dc_n]

    def recolour_mode(self):
        """Change the colour according to the system theme and active state."""
        if self.api_object is None:
            return

        if self.api_object.active:
            self.color = ACTIVE['color']
            self.style = ACTIVE['style']

        pen = QPen(self.color, self.pen_width, self.style)
        self.setPen(pen)

        self.terminal_ac.setPen(pen)
        self.terminal_dc_p.setPen(pen)
        self.terminal_dc_n.setPen(pen)

        if self.conn_line_ac is not None:
            self.conn_line_ac.recolour_mode()
        if self.conn_line_dc_p is not None:
            self.conn_line_dc_p.recolour_mode()
        if self.conn_line_dc_n is not None:
            self.conn_line_dc_n.recolour_mode()

        self.update()  # Request repaint

    def set_enable(self, val=True):
        """Set the enable value, graphically and in the API."""
        self.api_object.active = val
        self.recolour_mode()

    def mouseMoveEvent(self, event: 'QGraphicsSceneMouseEvent'):
        """Handle mouse move event."""
        super().mouseMoveEvent(event)
        self.editor.update_diagram_element(device=self.api_object,
                                           x=self.pos().x(),
                                           y=self.pos().y(),
                                           w=self.w,
                                           h=self.h,
                                           r=self.rotation(),
                                           graphic_object=self)
        self.update_conn()

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent):
        """Handle mouse press: select item and show in editor."""
        super().mousePressEvent(event)  # Handle selection
        if self.api_object is not None:
            self.editor.set_editor_model(api_object=self.api_object)

    def mouseDoubleClickEvent(self, event):
        """ On double-click, potentially open an editor (optional)."""
        # Currently no specific editor for VSC, but could be added.
        # self.edit()
        pass

    def delete(self):
        """Delete the VSC and its connections."""
        # Remove connections first
        self.remove_connection(self.conn_line_ac)
        self.remove_connection(self.conn_line_dc_p)
        self.remove_connection(self.conn_line_dc_n)

        # Remove the item itself
        self.editor.diagram_scene.removeItem(self)

    def set_connection(self, terminal_type: TerminalType, bus: Bus, conn_line: LineGraphicTemplateItem):
        """Set a connection to a specific terminal."""

        success = False
        # Check type compatibility and update API / store line
        if terminal_type == TerminalType.OTHER:
            print(f"Error: Invalid terminal type {terminal_type} for VSC {self.api_object.name}")

        elif terminal_type == TerminalType.AC:
            if bus.is_dc:
                self.editor.gui.show_error_toast(
                    f"Connecting AC terminal of VSC '{self.api_object.name}' to DC bus '{bus.name}'")
            elif self.conn_line_ac is not None:
                self.editor.gui.show_error_toast(f"AC terminal of VSC {self.api_object.name} is already connected.")
            else:
                self.api_object.bus_to = bus
                self.conn_line_ac = conn_line
                self.set_terminal_tooltips()
                success = True

        elif terminal_type == TerminalType.DC_P:
            if not bus.is_dc:
                self.editor.gui.show_error_toast(
                    f"Connecting DC+ terminal of VSC '{self.api_object.name}' to AC bus '{bus.name}'")
            elif self.conn_line_dc_p is not None:
                self.editor.gui.show_error_toast(f"AC terminal of VSC {self.api_object.name} is already connected.")
            else:
                self.api_object.bus_from = bus
                self.conn_line_dc_p = conn_line
                self.set_terminal_tooltips()
                success = True

        elif terminal_type == TerminalType.DC_N:
            if not bus.is_dc:
                self.editor.gui.show_error_toast(
                    f"Connecting DC- terminal of VSC '{self.api_object.name}' to AC bus '{bus.name}'")
            elif self.conn_line_dc_n is not None:
                self.editor.gui.show_error_toast(f"DC- terminal of VSC {self.api_object.name} is already connected.")
            else:
                self.api_object.bus_dc_n = bus
                self.conn_line_dc_n = conn_line
                self.set_terminal_tooltips()
                success = True

        return success

    def redraw(self):
        """
        Redraw method - only recalculate position if VSC doesn't have saved coordinates
        :return:
        """
        # Only recalculate position if the VSC is at the default position (close to 0, 0)
        # This preserves saved positions when loading from file
        current_pos = self.pos()
        # Use small tolerance to handle floating-point precision issues
        if abs(current_pos.x()) < 1.0 and abs(current_pos.y()) < 1.0:
            # Only do automatic positioning if we're near the origin
            h1 = self.terminal_ac.pos().y() - (self.terminal_dc_p.pos().y() + self.terminal_dc_n.pos().y()) / 2.0
            b1 = self.terminal_ac.pos().x() - (self.terminal_dc_p.pos().x() + self.terminal_dc_n.pos().x()) / 2.0
            ang = np.arctan2(h1, b1)
            h2 = self.rect().height() / 2.0
            w2 = self.rect().width() / 2.0
            a = h2 * np.cos(ang) - w2 * np.sin(ang)
            b = w2 * np.sin(ang) + h2 * np.cos(ang)

            center = (self.terminal_ac.pos() + (
                        self.terminal_dc_p.pos() + self.terminal_dc_n.pos()) * 0.5) * 0.5 - QPointF(a, b)

            transform = QTransform()
            transform.translate(center.x(), center.y())
            transform.rotate(np.rad2deg(ang))
            self.setTransform(transform)
        # If we're not near the origin, preserve the current position (likely loaded from file)

    def remove_connection(self, conn_line: LineGraphicTemplateItem):
        """Remove a connection from a specific terminal."""
        if conn_line is None:
            return
        else:
            line = conn_line
            # Remove the line from the scene
            if line.scene():
                line.scene().removeItem(line)
            # Optionally clear the bus/cn reference in the api_object
            if line == self.conn_line_ac:
                self.api_object.bus_to = None
            elif line == self.conn_line_dc_p:
                self.api_object.bus_from = None
            elif line == self.conn_line_dc_n:
                self.api_object.bus_dc_n = None
            self.set_terminal_tooltips()

    def contextMenuEvent(self, event):
        """Show context menu."""
        if self.api_object is not None:
            menu = QMenu()

            # Enable/Disable
            pe = menu.addAction('Enable/Disable')
            pe_icon = QIcon()
            if self.api_object.active:
                pe_icon.addPixmap(QPixmap(":/Icons/icons/uncheck_all.png"))
            else:
                pe_icon.addPixmap(QPixmap(":/Icons/icons/check_all.png"))
            pe.setIcon(pe_icon)
            pe.triggered.connect(lambda: self.set_enable(not self.api_object.active))

            # Draw Labels (if applicable)
            # add_menu_entry(menu=menu,
            #                text="Draw labels",
            #                function_ptr=self.enable_disable_label_drawing, # Needs implementation if label drawing is kept
            #                checkeable=True,
            #                checked_value=self.draw_labels)

            # Delete
            menu.addSeparator()
            ra2 = menu.addAction('Delete')
            del_icon = QIcon()
            del_icon.addPixmap(QPixmap(":/Icons/icons/delete3.png"))
            ra2.setIcon(del_icon)
            ra2.triggered.connect(self.delete)

            # --- Add back original control/profile actions ---
            menu.addSeparator()

            add_menu_entry(menu=menu,
                           text="Control V AC from DC+",
                           function_ptr=self.control_v_from,
                           icon_path=":/Icons/icons/edit.png")

            add_menu_entry(menu=menu,
                           text="Control V AC from AC",
                           function_ptr=self.control_v_to,
                           icon_path=":/Icons/icons/edit.png")

            menu.addSeparator()

            ra6 = menu.addAction('Plot profiles')
            plot_icon = QIcon()
            plot_icon.addPixmap(QPixmap(":/Icons/icons/plot.png"))
            ra6.setIcon(plot_icon)
            ra6.triggered.connect(self.plot_profiles)

            ra4 = menu.addAction('Assign rate to profile')
            ra4_icon = QIcon()
            ra4_icon.addPixmap(QPixmap(":/Icons/icons/assign_to_profile.png"))
            ra4.setIcon(ra4_icon)
            ra4.triggered.connect(self.assign_rate_to_profile)

            ra5 = menu.addAction('Assign active state to profile')
            ra5_icon = QIcon()
            ra5_icon.addPixmap(QPixmap(":/Icons/icons/assign_to_profile.png"))
            ra5.setIcon(ra5_icon)
            ra5.triggered.connect(self.assign_status_to_profile)

            menu.exec_(event.screenPos())
        else:
            pass

    def control_v_from(self):
        """
        Set control mode to regulate DC voltage based on DC+ side (Interpretation).
        Sets control1 to Vm_dcp.
        """
        if self.api_object.bus_to and self.api_object.bus_from:
            self.api_object.control1 = ConverterControlType.Vm_dc
            self.api_object.control1_dev = self.api_object.bus_from
            self.api_object.control1_val = 1.0  # Default to 1.0 pu

            print(f"VSC {self.api_object.name} control set: Control1=Vm_dcp (Bus: {self.api_object.bus_from.name})")
            self.editor.set_editor_model(api_object=self.api_object)  # Refresh editor view
        else:
            print("Error: Cannot set control_v_from, DC+ bus not connected.")

    def control_v_to(self):
        """
        Set control mode to regulate AC voltage based on AC side itself.
        Sets control1 to Vm_ac (controlling AC bus).
        Leaves control2 as is, or sets it to something default like Pac if necessary.
        """
        if self.api_object.bus_to:
            self.api_object.control1 = ConverterControlType.Vm_ac
            self.api_object.control1_dev = self.api_object.bus_to
            self.api_object.control1_val = 1.0  # Default to 1.0 pu

            print(f"VSC {self.api_object.name} control set: Control1=Vm_ac (Bus: {self.api_object.bus_to.name})")
            self.editor.set_editor_model(api_object=self.api_object)  # Refresh editor view
        else:
            print("Error: Cannot set control_v_to, AC bus not connected.")

    def assign_rate_to_profile(self):
        """
        Assign the snapshot rate to the profile
        """
        self._editor.set_rate_to_profile(self.api_object)

    def assign_status_to_profile(self):
        """
        Assign the snapshot rate to the profile
        """
        self._editor.set_active_status_to_profile(self.api_object)

    def plot_profiles(self) -> None:
        """
        Plot the time series profiles
        @return:
        """
        pass

    def assign_bus_to_vsc(self, terminal_vsc: RoundTerminalItem, bus_vsc:  RoundTerminalItem | BarTerminalItem) -> bool:
        """
        Assign the connected bus to a three-terminal VSC
        :return: if the connection was successful
        """
        if terminal_vsc.terminal_type == TerminalType.AC:
            if not bus_vsc.parent.api_object.is_dc:
                self.api_object.bus_to = bus_vsc.parent.api_object
                return True
            else:
                return False

        elif terminal_vsc.terminal_type == TerminalType.DC_P:
            if bus_vsc.parent.api_object.is_dc:
                self.api_object.bus_from = bus_vsc.parent.api_object
                return True
            else:
                return False

        elif terminal_vsc.terminal_type == TerminalType.DC_N:
            if bus_vsc.parent.api_object.is_dc:
                self.api_object.bus_dc_n = bus_vsc.parent.api_object
                return True
            else:
                return False

        else:
            print('Error in the VSC connection!')
            return False



# MIT License
#
# Copyright (c) 2021-2022 Yunosuke Ohsugi
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Utility methods for qdarktheme."""
from __future__ import annotations

import inspect
import logging
import operator as ope
import re
from pathlib import Path

import VeraGrid.ThirdParty.qdarktheme as qdarktheme

# greater_equal and less_equal must be evaluated before greater and less.
_OPERATORS = {"==": ope.eq, "!=": ope.ne, ">=": ope.ge, "<=": ope.le, ">": ope.gt, "<": ope.lt}


def multi_replace(target: str, replacements: dict[str, str]) -> str:
    """Given a string and a replacement map, it returns the replaced string.

    See https://gist.github.com/bgusach/a967e0587d6e01e889fd1d776c5f3729.

    Args:
        target: String to execute replacements on.
        replacements: Replacement dictionary {value to find: value to replace}.

    Returns:
        str: Target string that replaced with `replacements`.
    """
    if len(replacements) == 0:
        return target

    replacements_sorted = sorted(replacements, key=len, reverse=True)
    replacements_escaped = [re.escape(i) for i in replacements_sorted]
    pattern = re.compile("|".join(replacements_escaped))
    return pattern.sub(lambda match: replacements[match.group()], target)


def get_logger(logger_name: str) -> logging.Logger:
    """Return the logger with the name specified by logger_name arg.

    Args:
        logger_name: The name of logger.

    Returns:
        Logger reformatted for this package.
    """
    logger = logging.getLogger(logger_name)
    logger.propagate = False
    logger.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    ch.setFormatter(logging.Formatter("[%(name)s] [%(levelname)s] %(message)s"))
    logger.addHandler(ch)
    return logger


def get_cash_root_path(version: str) -> Path:
    """Return the cash root dir path."""
    return Path.home() / ".cache" / "qdarktheme" / f"v{version}"


def get_qdarktheme_root_path() -> Path:
    """Return the qdarktheme package root path.

    Returns:
        qdarktheme package root path.
    """
    return Path(inspect.getfile(qdarktheme)).parent


def _compare_v(v1: str, operator: str, v2: str) -> bool:
    """Comparing two versions."""
    v1_list, v2_list = (tuple(map(int, (v.split(".")))) for v in (v1, v2))
    return _OPERATORS[operator](v1_list, v2_list)


def analyze_version_str(target_version: str, version_text: str) -> bool:
    """Analyze text comparing versions."""
    for operator in _OPERATORS:
        if operator not in version_text:
            continue
        version = version_text.replace(operator, "")
        return _compare_v(target_version, operator, version)
    raise AssertionError("Text comparing versions is wrong.")



# MIT License
#
# Copyright (c) 2018 Ross Wilson
# Copyright (c) 2024, Santiago Peñate Vera
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
A base Tiles object for pySlipQt local tiles.

All tile sources should inherit from this base class.
For example, see gmt_local.py (local tiles) and osm_tiles.py
(internet tiles).
"""

import os
from typing import Tuple
from PySide6.QtGui import QPixmap
import VeraGrid.Gui.Diagrams.MapWidget.Tiles.pycacheback as pycacheback


class TilesCache(pycacheback.PyCacheBack):
    """Cache for local or internet tiles.

    Instance variables we use from pyCacheBack:
        self._tiles_dir  path to the on-disk cache directory
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.PicExtension = 'png'
        self.TilePath = '{Z}/{X}/{Y}.%s' % self.PicExtension

    def tile_date(self, key: Tuple[int, float, float]):
        """Return the creation date of a tile given its key."""

        tile_path = self.tile_path(key)
        return os.path.getctime(tile_path)

    def tile_path(self, key: Tuple[int, float, float]) -> str:
        """Return path to a tile file given its key."""

        (level, x, y) = key
        file_path = os.path.join(self._tiles_dir, self.TilePath.format(Z=level, X=x, Y=y))
        return file_path

    def _get_from_back(self, key: Tuple[int, float, float]) -> QPixmap:
        """Retrieve value for 'key' from backing storage.

        key  tuple (level, x, y)
             where level is the level of the tile
                   x, y  is the tile coordinates (integer)

        Raises KeyError if tile not found.
        """

        # look for item in disk cache
        file_path = self.tile_path(key)
        if not os.path.exists(file_path):
            # tile not there, raise KeyError
            raise KeyError("Item with key '%s' not found in on-disk cache"
                           % str(key)) from None

        # we have the tile file - read into memory & return
        return QPixmap(file_path)

    def _put_to_back(self, key: Tuple[int, float, float], image: QPixmap):
        """Put a image into on-disk cache.

        key     a tuple: (level, x, y)
                where level  level for image
                      x      integer tile coordinate
                      y      integer tile coordinate
        image   the wx.Image to save
        """

        (level, x, y) = key
        tile_path = os.path.join(self._tiles_dir,
                                 self.TilePath.format(Z=level, X=x, Y=y))
        dir_path = os.path.dirname(tile_path)
        try:
            os.makedirs(dir_path)
        except OSError:
            # we assume it's a 'directory exists' error, which we ignore
            pass

        image.save(tile_path, self.PicExtension)

    def add(self, key: Tuple[int, float, float], image: QPixmap):
        """
        Add entry
        :param key: key
        :param image: value
        """
        self._put_to_back(key, image)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'system_scaler_ui.ui'
##
## Created by: Qt User Interface Compiler version 6.9.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QDialog,
    QFrame, QHBoxLayout, QHeaderView, QLabel,
    QPushButton, QSizePolicy, QSpacerItem, QTableView,
    QVBoxLayout, QWidget)
from VeraGrid.Gui.Icons.icons_rc import *

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(525, 326)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(1, 1, 1, 1)
        self.frame = QFrame(Dialog)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.NoFrame)
        self.frame.setFrameShadow(QFrame.Raised)
        self.verticalLayout_2 = QVBoxLayout(self.frame)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.frame_2 = QFrame(self.frame)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setFrameShape(QFrame.NoFrame)
        self.frame_2.setFrameShadow(QFrame.Raised)
        self.horizontalLayout = QHBoxLayout(self.frame_2)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.group_combo_box = QComboBox(self.frame_2)
        self.group_combo_box.setObjectName(u"group_combo_box")

        self.horizontalLayout.addWidget(self.group_combo_box)

        self.label = QLabel(self.frame_2)
        self.label.setObjectName(u"label")

        self.horizontalLayout.addWidget(self.label)

        self.horizontalSpacer = QSpacerItem(340, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)


        self.verticalLayout_2.addWidget(self.frame_2)

        self.groups_scaling_table_view = QTableView(self.frame)
        self.groups_scaling_table_view.setObjectName(u"groups_scaling_table_view")

        self.verticalLayout_2.addWidget(self.groups_scaling_table_view)

        self.frame_3 = QFrame(self.frame)
        self.frame_3.setObjectName(u"frame_3")
        self.frame_3.setFrameShape(QFrame.NoFrame)
        self.frame_3.setFrameShadow(QFrame.Raised)
        self.horizontalLayout_2 = QHBoxLayout(self.frame_3)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.aply_to_time_series_check_box = QCheckBox(self.frame_3)
        self.aply_to_time_series_check_box.setObjectName(u"aply_to_time_series_check_box")

        self.horizontalLayout_2.addWidget(self.aply_to_time_series_check_box)

        self.horizontalSpacer_2 = QSpacerItem(250, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_2)

        self.doit_button = QPushButton(self.frame_3)
        self.doit_button.setObjectName(u"doit_button")

        self.horizontalLayout_2.addWidget(self.doit_button)


        self.verticalLayout_2.addWidget(self.frame_3)


        self.verticalLayout.addWidget(self.frame)


        self.retranslateUi(Dialog)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"System Scaler", None))
        self.label.setText(QCoreApplication.translate("Dialog", u"Grouping", None))
        self.aply_to_time_series_check_box.setText(QCoreApplication.translate("Dialog", u"Apply to time series", None))
        self.doit_button.setText(QCoreApplication.translate("Dialog", u"Do it", None))
    # retranslateUi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import PySide6  # this line is necessary so that Matplotlib recognises that PySide is the Qt Backend
import matplotlib
matplotlib.use('Qt6Agg')
from matplotlib import pyplot as plt  # leave here


########################################################################################################################
# Set Matplotlib global parameters
########################################################################################################################
# if 'fivethirtyeight' in plt.style.available:
#     plt.style.use('fivethirtyeight')

SMALL_SIZE = 8
MEDIUM_SIZE = 10
BIGGER_SIZE = 12
LINEWIDTH = 1

LEFT = 0.12
RIGHT = 0.98
TOP = 0.8
BOTTOM = 0.2
matplotlib.rc('font', size=SMALL_SIZE)  # controls default text sizes
matplotlib.rc('axes', titlesize=SMALL_SIZE)  # font size of the axes title
matplotlib.rc('axes', labelsize=SMALL_SIZE)  # font size of the x and y labels
matplotlib.rc('xtick', labelsize=SMALL_SIZE)  # font size of the tick labels
matplotlib.rc('ytick', labelsize=SMALL_SIZE)  # font size of the tick labels
matplotlib.rc('legend', fontsize=SMALL_SIZE)  # legend font size
matplotlib.rc('figure', titlesize=MEDIUM_SIZE)  # font size of the figure title



# MIT License
#
# Copyright (c) 2021-2022 Yunosuke Ohsugi
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import json
import re
from functools import lru_cache

from VeraGrid.ThirdParty.qdarktheme import _resources
from VeraGrid.ThirdParty.qdarktheme._color import Color


@lru_cache()
def _svg_resources() -> dict[str, str]:
    return json.loads(_resources.svg.SVG_RESOURCES)


class Svg:
    """Class to manage SVG."""

    _SVG_FILL_RE = re.compile(r'fill=".*?"')
    _SVG_FILL_OPACITY_RE = re.compile(r'fill-opacity=".*?"')
    _SVG_TRANSFORM_RE = re.compile(r'transform=".*?"')

    def __init__(self, id: str) -> None:
        """Initialize svg manager."""
        self._id = id
        self._color = None
        self._rotate = None
        self._source = _svg_resources()[self._id]

    def __str__(self) -> str:
        """Return the svg source code."""
        return self._source

    def colored(self, color: Color) -> Svg:
        """Add or change svg color."""
        svg_tiny_color_formats = color.to_svg_tiny_color_format().split(" ")
        if len(svg_tiny_color_formats) == 2:
            new_svg_color, new_svg_opacity = svg_tiny_color_formats
        else:
            new_svg_color = svg_tiny_color_formats[0]
            new_svg_opacity = None

        current_svg_color = Svg._SVG_FILL_RE.search(self._source)
        current_svg_opacity = Svg._SVG_FILL_OPACITY_RE.search(self._source)

        # Add or change SVG color.
        if current_svg_color is None:
            self._source = self._source.replace("<svg ", f"<svg {new_svg_color} ")
        else:
            self._source = self._source.replace(current_svg_color.group(), new_svg_color)

        # Add or change SVG opacity.
        if new_svg_opacity is not None and current_svg_opacity is None:
            self._source = self._source.replace("<svg ", f"<svg {new_svg_opacity} ")
        elif new_svg_opacity is not None and current_svg_opacity is not None:
            self._source = self._source.replace(current_svg_opacity.group(), new_svg_opacity)

        # Remove SVG opacity
        if new_svg_opacity is None and current_svg_opacity is not None:
            self._source = self._source.replace(" " + current_svg_opacity.group(), "")
        return self

    def rotate(self, rotate: int) -> Svg:
        """Rotate svg."""
        if rotate == 0:
            return self

        current_svg_transform = Svg._SVG_TRANSFORM_RE.search(self._source)
        new_svg_transform = f'transform="rotate({rotate}, 12, 12)"'
        if current_svg_transform is None:
            self._source = self._source.replace("<svg ", f"<svg {new_svg_transform} ")
        else:
            self._source = self._source.replace(current_svg_transform.group(), new_svg_transform)

        return self



# MIT License
#
# Copyright (c) 2021-2022 Yunosuke Ohsugi
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for color code."""
from __future__ import annotations

import colorsys
import math


def _round_float(number: float, decimal_points: int = 3) -> float:
    decimal = 10**decimal_points
    return round(number * decimal) / decimal


class _RGBA:
    """Class handling RGBA color code."""

    def __init__(self, r: int, g: int, b: int, a: float = 1) -> None:
        """Initialize rgba value.

        Args:
            r: Red(0~255).
            g: Green(0~255).
            b: Blue(0~255).
            a: Alpha(0~1). Defaults to 1.
        """
        self._r = min(255, max(0, r)) | 0
        self._g = min(255, max(0, g)) | 0
        self._b = min(255, max(0, b)) | 0
        self._a = _round_float(max(min(1, a), 0))

    def __str__(self) -> str:
        """Format RGBA class.

        e.g. rgba(100, 100, 100, 0.5).
        """
        return f"rgba({self.r}, {self.g}, {self.b}, {self.a:.3f})"

    def __getitem__(self, item: int) -> int | float:
        """Unpack to (r, g, b, a)."""
        return [self.r, self.g, self.b, self.a][item]

    def __eq__(self, other: _RGBA) -> bool:
        """Returns true if `r`, `g`, `b` and `a` are all the same."""
        return [self.r, self.g, self.b, self.a] == [other.r, other.g, other.b, other.a]

    @property
    def r(self) -> int:
        return self._r

    @property
    def g(self) -> int:
        return self._g

    @property
    def b(self) -> int:
        return self._b

    @property
    def a(self) -> float:
        return self._a


class _HSLA:
    def __init__(self, h: int, s: float, l: float, a: float = 1) -> None:  # noqa: E741
        self._h = max(min(360, h), 0) | 0
        self._s = _round_float(max(min(1, s), 0))
        self._l = _round_float(max(min(1, l), 0))
        self._a = _round_float(max(min(1, a), 0))

    def __eq__(self, other: _HSLA) -> bool:
        """Returns true if `h`, `s`, `l` and `a` are all the same."""
        return [self.h, self.s, self.l, self.a] == [other.h, other.s, other.l, other.a]

    @property
    def h(self) -> int:
        return self._h

    @property
    def s(self) -> float:
        return self._s

    @property
    def l(self) -> float:  # noqa: E741, E743
        return self._l

    @property
    def a(self) -> float:
        return self._a

    @staticmethod
    def from_rgba(rgba: _RGBA) -> _HSLA:
        hls = colorsys.rgb_to_hls(rgba.r / 255, rgba.g / 255, rgba.b / 255)
        return _HSLA(int(hls[0] * 360), hls[2], hls[1], rgba.a)

    def to_rgba(self) -> _RGBA:
        rgb = colorsys.hls_to_rgb(self.h / 360, self.l, self.s)
        return _RGBA(round(rgb[0] * 255), round(rgb[1] * 255), round(rgb[2] * 255), self.a)


class Color:
    """Class handling color code(RGBA and HSLA)."""

    def __init__(self, color_code: _RGBA | _HSLA) -> None:
        """Initialize color code."""
        self._hsla, self._hsva = None, None
        if isinstance(color_code, _RGBA):
            self._rgba = color_code
        elif isinstance(color_code, _HSLA):
            self._hsla = color_code
            self._rgba = self._hsla.to_rgba()

    @property
    def rgba(self) -> _RGBA:
        """Return rgba."""
        return self._rgba

    @property
    def hsla(self) -> _HSLA:
        """Return hsla."""
        return self._hsla if self._hsla else _HSLA.from_rgba(self.rgba)

    def __str__(self) -> str:
        """Format Color class.

        e.g. rgba(100, 100, 100, 0.5).
        """
        return str(self.rgba)

    @staticmethod
    def _check_hex_format(hex_format: str) -> None:
        """Check if string is hex format."""
        try:
            hex = hex_format.lstrip("#")
            if not len(hex) in (3, 4, 6, 8):
                raise ValueError
            int(hex, 16)
        except ValueError:
            raise ValueError(
                f'invalid hex color format: "{hex_format}". '
                "Only support following hexadecimal notations: #RGB, #RGBA, #RRGGBB and #RRGGBBAA. "
                "R (red), G (green), B (blue), and A (alpha) are hexadecimal characters "
                "(0-9, a-f or A-F)."
            ) from None

    @staticmethod
    def from_rgba(r: int, g: int, b: int, a: int) -> Color:
        """Convert rgba to Color object."""
        rgba = _RGBA(r, g, b, a / 255)
        return Color(rgba)

    @staticmethod
    def from_hex(val: str) -> Color:
        """
        Convert hex string to Color object.
        :param val: Color hex string.
        :return: Color object converted from hex.
        """
        Color._check_hex_format(val)
        val = val.lstrip("#")
        r, g, b, a = 255, 0, 0, 1
        if len(val) == 3:  # #RGB format
            r, g, b = (int(char, 16) for char in val)
            r, g, b = 16 * r + r, 16 * g + g, 16 * b + b
        if len(val) == 4:  # #RGBA format
            r, g, b, a = (int(char, 16) for char in val)
            r, g, b = 16 * r + r, 16 * g + g, 16 * b + b
            a = (16 * a + a) / 255
        if len(val) == 6:  # #RRGGBB format
            r, g, b = bytes.fromhex(val)
            a = 1
        elif len(val) == 8:  # #RRGGBBAA format
            r, g, b, a = bytes.fromhex(val)
            a = a / 255
        return Color(_RGBA(r, g, b, a))

    def _to_hex(self) -> str:
        """Convert Color object to hex(#RRGGBBAA).

        Args:
            color: Color object.

        Returns:
            str: Hex converted from Color object.
        """
        r, g, b, a = self.rgba.r, self.rgba.g, self.rgba.b, self.rgba.a
        hex_color = f"{math.floor(r):02x}{math.floor(g):02x}{math.floor(b):02x}"
        if a != 1:
            hex_color += f"{math.floor(a*255):02x}"
        return hex_color

    def to_hex_argb(self) -> str:
        """Convert Color object to hex(#AARRGGBB).

        Args:
            color: Color object.

        Returns:
            str: Hex converted from Color object.
        """
        r, g, b, a = self.rgba.r, self.rgba.g, self.rgba.b, self.rgba.a
        hex_color = "" if a == 1 else f"{math.floor(a*255):02x}"
        hex_color += f"{math.floor(r):02x}{math.floor(g):02x}{math.floor(b):02x}"
        return hex_color

    def to_svg_tiny_color_format(self) -> str:
        """Convert Color object to string for svg.

        QtSvg does not support #RRGGBBAA format.
        Therefore, we need to set the alpha value to `fill-opacity` instead.

        Returns:
            str: RGBA format.
        """
        r, g, b, a = self.rgba
        if a == 1:
            return f'fill="#{self._to_hex()}"'
        return f'fill="rgb({r},{g},{b})" fill-opacity="{a}"'

    def lighten(self, factor: float) -> Color:
        """Lighten color."""
        return Color(_HSLA(self.hsla.h, self.hsla.s, self.hsla.l + self.hsla.l * factor, self.hsla.a))

    def darken(self, factor: float) -> Color:
        """Darken color."""
        return Color(_HSLA(self.hsla.h, self.hsla.s, self.hsla.l - self.hsla.l * factor, self.hsla.a))

    def transparent(self, factor: float) -> Color:
        """Make color transparent."""
        return Color(_RGBA(self.rgba.r, self.rgba.g, self.rgba.b, self.rgba.a * factor))



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'grid_generator_gui.ui'
##
## Created by: Qt User Interface Compiler version 6.9.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QDoubleSpinBox, QFrame,
    QGridLayout, QHBoxLayout, QLabel, QPushButton,
    QSizePolicy, QSpacerItem, QSpinBox, QSplitter,
    QVBoxLayout, QWidget)

from VeraGrid.Gui.Widgets.matplotlibwidget import MatplotlibWidget
from VeraGrid.Gui.Icons.icons_rc import *

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(841, 518)
        icon = QIcon()
        icon.addFile(u":/Icons/icons/automatic_layout.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        MainWindow.setWindowIcon(icon)
        self.actionCopy_to_clipboard = QAction(MainWindow)
        self.actionCopy_to_clipboard.setObjectName(u"actionCopy_to_clipboard")
        icon1 = QIcon()
        icon1.addFile(u":/Icons/icons/copy.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionCopy_to_clipboard.setIcon(icon1)
        self.actionSave = QAction(MainWindow)
        self.actionSave.setObjectName(u"actionSave")
        icon2 = QIcon()
        icon2.addFile(u":/Icons/icons/import_profiles.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionSave.setIcon(icon2)
        self.verticalLayout_2 = QVBoxLayout(MainWindow)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.splitter_3 = QSplitter(MainWindow)
        self.splitter_3.setObjectName(u"splitter_3")
        self.splitter_3.setOrientation(Qt.Horizontal)
        self.frame_8 = QFrame(self.splitter_3)
        self.frame_8.setObjectName(u"frame_8")
        self.frame_8.setMaximumSize(QSize(400, 16777215))
        self.frame_8.setFrameShape(QFrame.NoFrame)
        self.frame_8.setFrameShadow(QFrame.Raised)
        self.verticalLayout_7 = QVBoxLayout(self.frame_8)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.verticalLayout_7.setContentsMargins(-1, 0, -1, -1)
        self.frame = QFrame(self.frame_8)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.NoFrame)
        self.frame.setFrameShadow(QFrame.Raised)
        self.gridLayout = QGridLayout(self.frame)
        self.gridLayout.setObjectName(u"gridLayout")
        self.power_SpinBox = QDoubleSpinBox(self.frame)
        self.power_SpinBox.setObjectName(u"power_SpinBox")
        self.power_SpinBox.setMaximum(9999999.000000000000000)
        self.power_SpinBox.setValue(100.000000000000000)

        self.gridLayout.addWidget(self.power_SpinBox, 7, 1, 1, 1)

        self.label_9 = QLabel(self.frame)
        self.label_9.setObjectName(u"label_9")
        palette = QPalette()
        brush = QBrush(QColor(136, 138, 133, 255))
        brush.setStyle(Qt.BrushStyle.SolidPattern)
        palette.setBrush(QPalette.ColorGroup.Active, QPalette.ColorRole.WindowText, brush)
        palette.setBrush(QPalette.ColorGroup.Active, QPalette.ColorRole.Text, brush)
        palette.setBrush(QPalette.ColorGroup.Active, QPalette.ColorRole.ButtonText, brush)
        brush1 = QBrush(QColor(136, 138, 133, 128))
        brush1.setStyle(Qt.BrushStyle.SolidPattern)
#if QT_VERSION >= QT_VERSION_CHECK(5, 12, 0)
        palette.setBrush(QPalette.ColorGroup.Active, QPalette.ColorRole.PlaceholderText, brush1)
#endif
        palette.setBrush(QPalette.ColorGroup.Inactive, QPalette.ColorRole.WindowText, brush)
        palette.setBrush(QPalette.ColorGroup.Inactive, QPalette.ColorRole.Text, brush)
        palette.setBrush(QPalette.ColorGroup.Inactive, QPalette.ColorRole.ButtonText, brush)
#if QT_VERSION >= QT_VERSION_CHECK(5, 12, 0)
        palette.setBrush(QPalette.ColorGroup.Inactive, QPalette.ColorRole.PlaceholderText, brush1)
#endif
        brush2 = QBrush(QColor(190, 190, 190, 255))
        brush2.setStyle(Qt.BrushStyle.SolidPattern)
        palette.setBrush(QPalette.ColorGroup.Disabled, QPalette.ColorRole.WindowText, brush2)
        palette.setBrush(QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text, brush2)
        palette.setBrush(QPalette.ColorGroup.Disabled, QPalette.ColorRole.ButtonText, brush2)
        brush3 = QBrush(QColor(0, 0, 0, 128))
        brush3.setStyle(Qt.BrushStyle.SolidPattern)
#if QT_VERSION >= QT_VERSION_CHECK(5, 12, 0)
        palette.setBrush(QPalette.ColorGroup.Disabled, QPalette.ColorRole.PlaceholderText, brush3)
#endif
        self.label_9.setPalette(palette)
        font = QFont()
        font.setPointSize(16)
        self.label_9.setFont(font)

        self.gridLayout.addWidget(self.label_9, 0, 0, 1, 1)

        self.b_SpinBox = QDoubleSpinBox(self.frame)
        self.b_SpinBox.setObjectName(u"b_SpinBox")
        self.b_SpinBox.setDecimals(6)
        self.b_SpinBox.setValue(0.000010000000000)

        self.gridLayout.addWidget(self.b_SpinBox, 10, 1, 1, 1)

        self.label_2 = QLabel(self.frame)
        self.label_2.setObjectName(u"label_2")

        self.gridLayout.addWidget(self.label_2, 7, 0, 1, 1)

        self.r_SpinBox = QDoubleSpinBox(self.frame)
        self.r_SpinBox.setObjectName(u"r_SpinBox")
        self.r_SpinBox.setDecimals(6)
        self.r_SpinBox.setValue(0.000100000000000)

        self.gridLayout.addWidget(self.r_SpinBox, 8, 1, 1, 1)

        self.label_3 = QLabel(self.frame)
        self.label_3.setObjectName(u"label_3")

        self.gridLayout.addWidget(self.label_3, 1, 0, 1, 1)

        self.label_6 = QLabel(self.frame)
        self.label_6.setObjectName(u"label_6")

        self.gridLayout.addWidget(self.label_6, 10, 0, 1, 1)

        self.x_SpinBox = QDoubleSpinBox(self.frame)
        self.x_SpinBox.setObjectName(u"x_SpinBox")
        self.x_SpinBox.setDecimals(6)
        self.x_SpinBox.setValue(0.001000000000000)

        self.gridLayout.addWidget(self.x_SpinBox, 9, 1, 1, 1)

        self.load_nodes_SpinBox = QDoubleSpinBox(self.frame)
        self.load_nodes_SpinBox.setObjectName(u"load_nodes_SpinBox")
        self.load_nodes_SpinBox.setDecimals(1)
        self.load_nodes_SpinBox.setMinimum(1.000000000000000)
        self.load_nodes_SpinBox.setMaximum(100.000000000000000)
        self.load_nodes_SpinBox.setValue(50.000000000000000)

        self.gridLayout.addWidget(self.load_nodes_SpinBox, 5, 1, 1, 1)

        self.label_4 = QLabel(self.frame)
        self.label_4.setObjectName(u"label_4")

        self.gridLayout.addWidget(self.label_4, 8, 0, 1, 1)

        self.ratio_SpinBox = QDoubleSpinBox(self.frame)
        self.ratio_SpinBox.setObjectName(u"ratio_SpinBox")
        self.ratio_SpinBox.setDecimals(4)
        self.ratio_SpinBox.setMinimum(0.000000000000000)
        self.ratio_SpinBox.setMaximum(1.000000000000000)
        self.ratio_SpinBox.setValue(0.333000000000000)

        self.gridLayout.addWidget(self.ratio_SpinBox, 2, 1, 1, 1)

        self.label_11 = QLabel(self.frame)
        self.label_11.setObjectName(u"label_11")

        self.gridLayout.addWidget(self.label_11, 2, 0, 1, 1)

        self.nodes_spinBox = QSpinBox(self.frame)
        self.nodes_spinBox.setObjectName(u"nodes_spinBox")
        self.nodes_spinBox.setMinimum(2)
        self.nodes_spinBox.setMaximum(999999999)
        self.nodes_spinBox.setValue(100)

        self.gridLayout.addWidget(self.nodes_spinBox, 1, 1, 1, 1)

        self.label_10 = QLabel(self.frame)
        self.label_10.setObjectName(u"label_10")
        palette1 = QPalette()
        palette1.setBrush(QPalette.ColorGroup.Active, QPalette.ColorRole.WindowText, brush)
        palette1.setBrush(QPalette.ColorGroup.Active, QPalette.ColorRole.Text, brush)
        palette1.setBrush(QPalette.ColorGroup.Active, QPalette.ColorRole.ButtonText, brush)
#if QT_VERSION >= QT_VERSION_CHECK(5, 12, 0)
        palette1.setBrush(QPalette.ColorGroup.Active, QPalette.ColorRole.PlaceholderText, brush1)
#endif
        palette1.setBrush(QPalette.ColorGroup.Inactive, QPalette.ColorRole.WindowText, brush)
        palette1.setBrush(QPalette.ColorGroup.Inactive, QPalette.ColorRole.Text, brush)
        palette1.setBrush(QPalette.ColorGroup.Inactive, QPalette.ColorRole.ButtonText, brush)
#if QT_VERSION >= QT_VERSION_CHECK(5, 12, 0)
        palette1.setBrush(QPalette.ColorGroup.Inactive, QPalette.ColorRole.PlaceholderText, brush1)
#endif
        palette1.setBrush(QPalette.ColorGroup.Disabled, QPalette.ColorRole.WindowText, brush2)
        palette1.setBrush(QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text, brush2)
        palette1.setBrush(QPalette.ColorGroup.Disabled, QPalette.ColorRole.ButtonText, brush2)
#if QT_VERSION >= QT_VERSION_CHECK(5, 12, 0)
        palette1.setBrush(QPalette.ColorGroup.Disabled, QPalette.ColorRole.PlaceholderText, brush3)
#endif
        self.label_10.setPalette(palette1)
        self.label_10.setFont(font)

        self.gridLayout.addWidget(self.label_10, 3, 0, 1, 1)

        self.label_8 = QLabel(self.frame)
        self.label_8.setObjectName(u"label_8")

        self.gridLayout.addWidget(self.label_8, 5, 0, 1, 1)

        self.generation_nodes_SpinBox = QDoubleSpinBox(self.frame)
        self.generation_nodes_SpinBox.setObjectName(u"generation_nodes_SpinBox")
        self.generation_nodes_SpinBox.setDecimals(1)
        self.generation_nodes_SpinBox.setMinimum(1.000000000000000)
        self.generation_nodes_SpinBox.setMaximum(100.000000000000000)
        self.generation_nodes_SpinBox.setValue(30.000000000000000)

        self.gridLayout.addWidget(self.generation_nodes_SpinBox, 4, 1, 1, 1)

        self.label_7 = QLabel(self.frame)
        self.label_7.setObjectName(u"label_7")

        self.gridLayout.addWidget(self.label_7, 4, 0, 1, 1)

        self.label_5 = QLabel(self.frame)
        self.label_5.setObjectName(u"label_5")

        self.gridLayout.addWidget(self.label_5, 9, 0, 1, 1)

        self.label_12 = QLabel(self.frame)
        self.label_12.setObjectName(u"label_12")

        self.gridLayout.addWidget(self.label_12, 6, 0, 1, 1)

        self.power_factor_SpinBox = QDoubleSpinBox(self.frame)
        self.power_factor_SpinBox.setObjectName(u"power_factor_SpinBox")
        self.power_factor_SpinBox.setMaximum(1.000000000000000)
        self.power_factor_SpinBox.setValue(0.800000000000000)

        self.gridLayout.addWidget(self.power_factor_SpinBox, 6, 1, 1, 1)


        self.verticalLayout_7.addWidget(self.frame)

        self.frame_2 = QFrame(self.frame_8)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setFrameShape(QFrame.NoFrame)
        self.frame_2.setFrameShadow(QFrame.Raised)
        self.horizontalLayout_2 = QHBoxLayout(self.frame_2)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.applyButton = QPushButton(self.frame_2)
        self.applyButton.setObjectName(u"applyButton")
        self.applyButton.setMinimumSize(QSize(24, 24))
        icon3 = QIcon()
        icon3.addFile(u":/Icons/icons/color_grid.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.applyButton.setIcon(icon3)

        self.horizontalLayout_2.addWidget(self.applyButton)

        self.previewButton = QPushButton(self.frame_2)
        self.previewButton.setObjectName(u"previewButton")
        self.previewButton.setMinimumSize(QSize(24, 24))
        icon4 = QIcon()
        icon4.addFile(u":/Icons/icons/run_cascade_step.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.previewButton.setIcon(icon4)

        self.horizontalLayout_2.addWidget(self.previewButton)

        self.horizontalSpacer = QSpacerItem(543, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer)


        self.verticalLayout_7.addWidget(self.frame_2)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_7.addItem(self.verticalSpacer)

        self.splitter_3.addWidget(self.frame_8)
        self.PlotFrame = QFrame(self.splitter_3)
        self.PlotFrame.setObjectName(u"PlotFrame")
        self.PlotFrame.setFrameShape(QFrame.NoFrame)
        self.PlotFrame.setFrameShadow(QFrame.Raised)
        self.horizontalLayout = QHBoxLayout(self.PlotFrame)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.plotwidget = MatplotlibWidget(self.PlotFrame)
        self.plotwidget.setObjectName(u"plotwidget")

        self.horizontalLayout.addWidget(self.plotwidget)

        self.splitter_3.addWidget(self.PlotFrame)

        self.verticalLayout.addWidget(self.splitter_3)


        self.verticalLayout_2.addLayout(self.verticalLayout)


        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"MainWindow", None))
        self.actionCopy_to_clipboard.setText(QCoreApplication.translate("MainWindow", u"Copy to clipboard", None))
        self.actionSave.setText(QCoreApplication.translate("MainWindow", u"Save", None))
        self.power_SpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" MW", None))
        self.label_9.setText(QCoreApplication.translate("MainWindow", u"Graph", None))
        self.b_SpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" p.u./km", None))
        self.label_2.setText(QCoreApplication.translate("MainWindow", u"Grid power", None))
        self.r_SpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" p.u./km", None))
        self.label_3.setText(QCoreApplication.translate("MainWindow", u"Number of nodes", None))
        self.label_6.setText(QCoreApplication.translate("MainWindow", u"Susceptance (B)", None))
        self.x_SpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" p.u./km", None))
        self.load_nodes_SpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" %", None))
        self.label_4.setText(QCoreApplication.translate("MainWindow", u"Resistance (R)", None))
        self.label_11.setText(QCoreApplication.translate("MainWindow", u"Expansion ratio", None))
        self.nodes_spinBox.setSuffix(QCoreApplication.translate("MainWindow", u" nodes", None))
        self.label_10.setText(QCoreApplication.translate("MainWindow", u"Devices", None))
        self.label_8.setText(QCoreApplication.translate("MainWindow", u"Load nodes", None))
        self.generation_nodes_SpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" %", None))
        self.label_7.setText(QCoreApplication.translate("MainWindow", u"Generation nodes", None))
        self.label_5.setText(QCoreApplication.translate("MainWindow", u"Reactance(X) ", None))
        self.label_12.setText(QCoreApplication.translate("MainWindow", u"Power factor", None))
#if QT_CONFIG(tooltip)
        self.applyButton.setToolTip(QCoreApplication.translate("MainWindow", u"Create Grid", None))
#endif // QT_CONFIG(tooltip)
        self.applyButton.setText("")
#if QT_CONFIG(tooltip)
        self.previewButton.setToolTip(QCoreApplication.translate("MainWindow", u"Preview", None))
#endif // QT_CONFIG(tooltip)
        self.previewButton.setText("")
    # retranslateUi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import os
from typing import Union, List, Callable
from PySide6 import QtWidgets

import VeraGrid.Gui.gui_functions as gf
import VeraGrid.Session.export_results_driver as exprtdrv
import VeraGrid.Session.file_handler as filedrv
from VeraGrid.Gui.GridMerge.grid_diff import GridDiffDialogue
from VeraGrid.Gui.GridMerge.grid_merge import GridMergeDialogue
from VeraGrid.plugins import install_plugin, get_plugin_info
from VeraGrid.Gui.CoordinatesInput.coordinates_dialogue import CoordinatesInputGUI
from VeraGrid.Gui.general_dialogues import LogsDialogue, CustomQuestionDialogue
from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget
from VeraGrid.Gui.messages import yes_no_question, error_msg, warning_msg, info_msg
from VeraGrid.Gui.GridGenerator.grid_generator_dialogue import GridGeneratorGUI
from VeraGrid.Gui.RosetaExplorer.RosetaExplorer import RosetaExplorerGUI
from VeraGrid.Gui.Main.SubClasses.Settings.configuration import ConfigurationMain
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.Compilers.circuit_to_newton_pa import NEWTON_PA_AVAILABLE
from VeraGridEngine.Compilers.circuit_to_pgm import PGM_AVAILABLE
from VeraGridEngine.IO.cim.cgmes.cgmes_export import get_available_cgmes_profiles
from VeraGridEngine.enumerations import CGMESVersions, SimulationTypes
from VeraGridEngine.IO.veragrid.contingency_parser import import_contingencies_from_json, export_contingencies_json_file
from VeraGridEngine.IO.cim.cgmes.cgmes_enums import CgmesProfileType
from VeraGridEngine.IO.veragrid.remote import RemoteInstruction
from VeraGridEngine.IO.veragrid.catalogue import save_catalogue, load_catalogue
from VeraGridEngine.Utils.ThirdParty.gslv.gslv_activation import install_gslv_license
from VeraGrid.templates import (get_cables_catalogue, get_transformer_catalogue, get_wires_catalogue,
                                get_sequence_lines_catalogue, get_rms_model_catalogue)


class IoMain(ConfigurationMain):
    """
    Inputs-Outputs Main
    """

    def __init__(self, parent=None):
        """

        @param parent:
        """

        # create main window
        ConfigurationMain.__init__(self, parent)

        self.rosetta_gui: Union[RosetaExplorerGUI, None] = None

        self.accepted_extensions = ['.veragrid', '.dveragrid',
                                    '.gridcal', '.dgridcal',
                                    '.xlsx', '.xls', '.sqlite', '.gch5',
                                    '.dgs', '.m', '.raw', '.RAW', '.json', '.uct',
                                    '.ejson2', '.ejson3', '.p', '.nc', '.hdf5',
                                    '.xml', '.rawx', '.zip', '.dpx', '.pwf', '.epc', '.EPC',
                                    '.vgplugin', '.gslv']

        self.cgmes_version_dict = {x.value: x for x in [CGMESVersions.v2_4_15,
                                                        CGMESVersions.v3_0_0]}
        self.ui.cgmes_version_comboBox.setModel(gf.get_list_model(list(self.cgmes_version_dict.keys())))

        self.cgmes_profiles_dict = {key: CgmesProfileType(key) for key, val in
                                    get_available_cgmes_profiles(cgmes_version=CGMESVersions.v2_4_15).items()}

        self.ui.cgmes_profiles_listView.setModel(gf.get_list_model(list(self.cgmes_profiles_dict.keys()),
                                                                   checks=True, check_value=True))

        self.ui.raw_export_version_comboBox.addItems(["33", "35"])

        self.ui.actionNew_project.triggered.connect(self.new_project)
        self.ui.actionOpen_file.triggered.connect(self.open_file)
        self.ui.actionAdd_circuit.triggered.connect(self.import_circuit)
        self.ui.actionExport_circuit_differential.triggered.connect(self.export_circuit_differential)
        self.ui.actionSave.triggered.connect(self.save_file)
        self.ui.actionSave_as.triggered.connect(self.save_file_as)
        self.ui.actionExport_all_the_device_s_profiles.triggered.connect(self.export_object_profiles)
        self.ui.actionExport_all_results.triggered.connect(self.export_all)
        self.ui.actiongrid_Generator.triggered.connect(self.grid_generator)
        self.ui.actionImport_bus_coordinates.triggered.connect(self.import_bus_coordinates)
        self.ui.actionImport_contingencies.triggered.connect(self.import_contingencies)
        self.ui.actionExport_contingencies.triggered.connect(self.export_contingencies)
        self.ui.actionAdd_default_catalogue.triggered.connect(self.add_default_catalogue)
        self.ui.actionAdd_custom_catalogue.triggered.connect(self.load_custom_catalogue)
        self.ui.actionExportCatalogue.triggered.connect(self.save_custom_catalogue)

        # Buttons
        self.ui.exportSimulationDataButton.clicked.connect(self.export_simulation_data)
        self.ui.loadResultFromDiskButton.clicked.connect(self.load_results_driver)

        # change
        self.ui.cgmes_version_comboBox.currentTextChanged.connect(self.cgmes_version_change)

    def dragEnterEvent(self, event):
        """

        :param event:
        :return:
        """
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        """

        :param event:
        :return:
        """
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """
        Drop file on the GUI, the default behaviour is to load the file
        :param event: event containing all the information
        """
        if event.mimeData().hasUrls:
            events = event.mimeData().urls()
            if len(events) > 0:

                file_names = list()

                any_grid_delta = False
                any_normal_grid = False

                for event in events:
                    file_name = event.toLocalFile()
                    name, file_extension = os.path.splitext(file_name)
                    if file_extension.lower() in self.accepted_extensions:
                        file_names.append(file_name)

                        if file_name.endswith('.dgridcal') or file_name.endswith('.dveragrid'):
                            any_grid_delta = True
                        elif file_name.endswith('.vgplugin'):
                            self.install_plugin_now(file_name)
                            return
                        elif file_name.endswith('.license.gslv'):
                            ok, msg = install_gslv_license(file_name)
                            if ok:
                                self.show_info_toast(msg)
                            else:
                                self.show_error_toast(msg)
                            return
                        else:
                            any_normal_grid = True

                    else:
                        error_msg('The file type ' + file_extension.lower() + ' is not accepted :(')

                if self.circuit.valid_for_simulation() > 0:

                    if any_grid_delta and not any_normal_grid:
                        # only grid deltas...
                        self.open_file_now(filenames=file_names,
                                           post_function=self.post_import_circuit)
                    else:
                        quit_msg = ("Are you sure that you want to quit the current grid and open a new one?"
                                    "\n If the process is cancelled the grid will remain.")
                        reply = QtWidgets.QMessageBox.question(self, 'Message', quit_msg,
                                                               QtWidgets.QMessageBox.StandardButton.Yes,
                                                               QtWidgets.QMessageBox.StandardButton.No)

                        if reply == QtWidgets.QMessageBox.StandardButton.Yes.value:
                            self.open_file_now(filenames=file_names)
                else:
                    # Just open the file
                    self.open_file_now(filenames=file_names)

    def new_project_now(self, create_default_diagrams=True):
        """
        New project right now without asking questions
        """
        # clear the circuit model
        self.circuit = MultiCircuit()

        # clear the file name
        self.file_name = ''

        self.remove_all_diagrams()

        self.setup_objects_tree()

        # clear the results
        self.ui.resultsTableView.setModel(None)
        self.ui.resultsLogsTreeView.setModel(None)

        # clear the comments
        self.ui.comments_textEdit.setText("")

        self.ui.grid_name_line_edit.setText("")

        # clear the simulation objects
        for thread in self.get_all_threads():
            thread = None

        if self.analysis_dialogue is not None:
            self.analysis_dialogue.close()

        self.clear_stuff_running()
        self.clear_results()

        if create_default_diagrams:
            self.add_complete_bus_branch_diagram()
            self.add_map_diagram()
            self.set_diagram_widget(self.diagram_widgets_list[0])

        self.collect_memory()

    def new_project(self):
        """
        Create new grid
        :return:
        """
        if self.circuit.valid_for_simulation() > 0:
            quit_msg = "Are you sure that you want to quit the current grid and create a new one?"
            reply = QtWidgets.QMessageBox.question(self, 'Message', quit_msg,
                                                   QtWidgets.QMessageBox.StandardButton.Yes,
                                                   QtWidgets.QMessageBox.StandardButton.No)

            if reply == QtWidgets.QMessageBox.StandardButton.Yes.value:
                self.new_project_now(create_default_diagrams=True)

    def open_file(self):
        """
        Open VeraGrid file
        @return:
        """
        if ('file_save' not in self.stuff_running_now) and ('file_open' not in self.stuff_running_now):
            if self.circuit.valid_for_simulation() > 0:
                quit_msg = ("Are you sure that you want to quit the current grid and open a new one?"
                            "\n If the process is cancelled the grid will remain.")
                reply = QtWidgets.QMessageBox.question(self, 'Message', quit_msg,
                                                       QtWidgets.QMessageBox.StandardButton.Yes,
                                                       QtWidgets.QMessageBox.StandardButton.No)

                if reply == QtWidgets.QMessageBox.StandardButton.Yes.value:
                    self.open_file_threaded()
                else:
                    pass
            else:
                # Just open the file
                self.open_file_threaded()

        else:
            warning_msg('There is a file being processed now.')

    def open_file_threaded(self, post_function=None,
                           allow_diff_file_format: bool = False,
                           title: str = 'Open file'):
        """
        Open file from a Qt thread to remain responsive
        :param post_function: Any function to run after
        :param allow_diff_file_format: Allow loading VeraGrid diff files?
        :param title: Title of the open window
        """

        files_types = "*.gridcal *.veragrid "

        if allow_diff_file_format:
            files_types += "*.dgridcal *.dveragrid "

        files_types += "*.gch5 *.xlsx *.xls *.sqlite *.dgs "
        files_types += "*.m *.raw *.RAW *.rawx *.uct *.json *.ejson2 *.ejson3 *.xml "
        files_types += "*.zip *.dpx *.pwf *.epc *.EPC *.nc *.hdf5 *.p"

        dialogue = QtWidgets.QFileDialog(None,
                                         caption=title,
                                         directory=self.project_directory,
                                         filter=f"Formats ({files_types})")

        if dialogue.exec():
            filenames = dialogue.selectedFiles()
            self.open_file_now(filenames, post_function)

    def open_file_now(self, filenames: Union[str, List[str]],
                      post_function: Union[None, Callable[[], None]] = None) -> None:
        """
        Open a file without questions
        :param filenames: list of file names (maybe more than one because of CIM TP and EQ files)
        :param post_function: function callback
        :return: Nothing
        """
        if len(filenames) > 0:

            for f_name in filenames:
                if not os.path.exists(f_name):
                    error_msg(text=f"The file does not exists :(\n{f_name}", title="File opening")
                    return

            self.file_name = filenames[0]

            # store the working directory
            self.project_directory = os.path.dirname(self.file_name)

            # lock the ui
            self.LOCK()

            # create thread
            self.open_file_thread_object = filedrv.FileOpenThread(
                file_name=filenames if len(filenames) > 1 else filenames[0],
                previous_circuit=self.circuit,
                options=self.get_file_open_options()
            )

            # make connections
            self.open_file_thread_object.progress_signal.connect(self.ui.progressBar.setValue)
            self.open_file_thread_object.progress_text.connect(self.ui.progress_label.setText)
            self.open_file_thread_object.done_signal.connect(self.UNLOCK)

            if post_function is None:
                self.open_file_thread_object.done_signal.connect(self.post_open_file)
            else:
                self.open_file_thread_object.done_signal.connect(post_function)

            # thread start
            self.open_file_thread_object.start()

            # register as the latest file driver
            self.last_file_driver = self.open_file_thread_object

            # register thread
            self.stuff_running_now.append(SimulationTypes.FileOpen)

    def post_open_file(self) -> None:
        """
        Actions to perform after a file has been loaded
        """

        self.stuff_running_now.remove(SimulationTypes.FileOpen)

        if self.open_file_thread_object is not None:

            if self.open_file_thread_object.valid:

                # assign the loaded circuit
                self.new_project_now(create_default_diagrams=False)
                if self.open_file_thread_object.circuit is not None:
                    self.circuit = self.open_file_thread_object.circuit
                    self.file_name = self.open_file_thread_object.file_name

                if self.circuit.has_diagrams():
                    # create the diagrams that came with the file
                    self.create_circuit_stored_diagrams()

                else:
                    if self.circuit.get_bus_number() > 300:
                        self.show_info_toast("The grid is quite big, no diagram is automatically created",
                                             duration=5000)

                    else:
                        # create schematic
                        self.add_complete_bus_branch_diagram()

                # set base magnitudes
                self.ui.sbase_doubleSpinBox.setValue(self.circuit.Sbase)
                self.ui.fbase_doubleSpinBox.setValue(self.circuit.fBase)

                # set circuit comments
                try:
                    self.ui.comments_textEdit.setText(str(self.circuit.comments))
                except ValueError:
                    pass

                # update the drop-down menus that display dates
                self.update_date_dependent_combos()
                self.update_from_to_list_views()

                # get the session tree structure
                session_data_dict = self.open_file_thread_object.get_session_tree()
                mdl = gf.get_tree_model(session_data_dict, 'Sessions')
                self.ui.diskSessionsTreeView.setModel(mdl)

                # apply the GUI settings if found:
                gui_config_data = self.open_file_thread_object.json_files.get('gui_config', None)
                if gui_config_data is not None:
                    self.apply_gui_config(data=gui_config_data)

                # clear the results
                self.clear_results()

                self.ui.grid_name_line_edit.setText(self.circuit.name)

                # if this was a CGMES file, launch the Rosetta GUI
                if self.open_file_thread_object.cgmes_circuit:

                    show_rosetta = yes_no_question(title="Show rosetta",
                                                   text="Do you want to open the Rosetta CGMEs browser?")

                    if show_rosetta:
                        self.rosetta_gui = RosetaExplorerGUI()
                        self.rosetta_gui.set_grid_model(self.open_file_thread_object.cgmes_circuit)
                        self.rosetta_gui.set_logger(self.open_file_thread_object.cgmes_logger)
                        self.rosetta_gui.update_combo_boxes()
                        self.rosetta_gui.show()
                    else:
                        # else, show the logger if it is necessary
                        if len(self.open_file_thread_object.logger) > 0:
                            dlg = LogsDialogue('Open CGMES file logger', self.open_file_thread_object.logger)
                            dlg.exec()

                else:
                    # else, show the logger if it is necessary
                    if len(self.open_file_thread_object.logger) > 0:
                        dlg = LogsDialogue('Open file logger', self.open_file_thread_object.logger)
                        dlg.exec()

            else:
                warning_msg(text='Error while loading the file(s)')
                # else, show the logger if it is necessary
                if len(self.open_file_thread_object.logger) > 0:
                    dlg = LogsDialogue('Open file logger', self.open_file_thread_object.logger)
                    dlg.exec()
        else:
            # center nodes
            diagram = self.get_selected_diagram_widget()
            if diagram is not None:
                if isinstance(diagram, SchematicWidget):
                    diagram.center_nodes()

        self.collect_memory()
        self.setup_time_sliders()
        self.get_circuit_snapshot_datetime()
        self.change_theme_mode()

    def install_plugin_now(self, file_name: str):
        """
        Install plugin
        :param file_name: name of the plugin
        """
        if file_name.endswith('.vgplugin'):
            info = get_plugin_info(file_name)

            if info is not None:

                if not info.is_compatible():
                    error_msg(f"{info.name} {info.version} requires VeraGrid {info.veragrid_version}",
                              "Plugin install")
                    return

                # search for the plugin
                for key, plugin in self.plugins_info.plugins.items():
                    if plugin.name == info.name:

                        ok = yes_no_question(f"There is a plugin already: "
                                             f"{plugin.name} {plugin.version} "
                                             f"The new plugin is {info.version}. "
                                             f"Install?", "Plugin install")
                        if not ok:
                            return
                        else:
                            break

                install_plugin(file_name)
                self.add_plugins()
                info_msg(f"{info.name} {info.version} installed!", "Plugin install")
            else:
                error_msg("There is no manifest :(", "Plugin install")
        else:
            error_msg("Does not seem to be a plugin :/", "Plugin install")

    def select_csv_file(self, caption='Open CSV file'):
        """
        Select a CSV file
        :return: csv file path
        """
        files_types = "CSV (*.csv)"

        filename, type_selected = QtWidgets.QFileDialog.getOpenFileName(parent=self,
                                                                        caption=caption,
                                                                        dir=self.project_directory,
                                                                        filter=files_types)

        if len(filename) > 0:
            return filename
        else:
            return None

    def import_circuit(self):
        """
        Prompt to add another circuit
        """
        self.open_file_threaded(post_function=self.post_import_circuit,
                                allow_diff_file_format=True)

    def post_import_circuit(self):
        """
        Stuff to do after opening another circuit
        :return: Nothing
        """
        self.stuff_running_now.remove(SimulationTypes.FileOpen)

        if self.open_file_thread_object is not None:

            new_circuit = self.open_file_thread_object.circuit

            if len(self.open_file_thread_object.logger) > 0:
                dlg = LogsDialogue('Open file logger',
                                   self.open_file_thread_object.logger)
                dlg.exec()

            if self.open_file_thread_object.valid:

                merge_dlg = GridMergeDialogue(grid=self.circuit, diff=new_circuit)
                merge_dlg.exec_()

                if merge_dlg.added_grid:
                    # Create a blank diagram and add to it
                    # logger = self.circuit.add_circuit(new_circuit)

                    # for sure, we want to add to the current schematic
                    diagram_widget = self.get_selected_diagram_widget()

                elif merge_dlg.merged_grid:

                    dlg3 = CustomQuestionDialogue(title="Grid merge",
                                                  question="How do you want to represent the merged grid?",
                                                  answer1="Create new diagram",
                                                  answer2="Add to current diagram")
                    dlg3.exec()

                    if dlg3.accepted_answer == 1:
                        # Create a blank diagram and add to it
                        diagram_widget = self.create_blank_schematic_diagram(name=new_circuit.name)

                    elif dlg3.accepted_answer == 2:
                        diagram_widget = self.get_selected_diagram_widget()

                    else:
                        # not imported
                        return

                    if diagram_widget is not None:
                        if isinstance(diagram_widget, SchematicWidget):
                            injections_by_bus = new_circuit.get_injection_devices_grouped_by_bus()
                            injections_by_fluid_node = new_circuit.get_injection_devices_grouped_by_fluid_node()

                            diagram_widget.add_elements_to_schematic(buses=new_circuit.buses,
                                                                     lines=new_circuit.lines,
                                                                     dc_lines=new_circuit.dc_lines,
                                                                     transformers2w=new_circuit.transformers2w,
                                                                     transformers3w=new_circuit.transformers3w,
                                                                     hvdc_lines=new_circuit.hvdc_lines,
                                                                     vsc_devices=new_circuit.vsc_devices,
                                                                     upfc_devices=new_circuit.upfc_devices,
                                                                     switches=new_circuit.switch_devices,
                                                                     fluid_nodes=new_circuit.fluid_nodes,
                                                                     fluid_paths=new_circuit.fluid_paths,
                                                                     injections_by_bus=injections_by_bus,
                                                                     injections_by_fluid_node=injections_by_fluid_node,
                                                                     explode_factor=1.0,
                                                                     prog_func=None,
                                                                     text_func=None)
                            diagram_widget.set_selected_buses(buses=new_circuit.buses)
                        else:
                            info_msg("No schematic diagram was selected...", title="Add to current diagram")

                else:
                    return

    def export_circuit_differential(self):
        """
        Prompt to export a diff of this circuit and a base one
        """
        dlg = GridDiffDialogue(grid=self.circuit)
        dlg.exec()

    def save_file_as(self):
        """
        Save this file as...
        """
        # by deleting the file_name, the save_file function will ask for it
        self.file_name = ''
        self.save_file()

    def save_file(self):
        """
        Save the circuit case to a file
        """

        if self.server_driver.is_running():
            instruction = RemoteInstruction(operation=SimulationTypes.NoSim)
            self.server_driver.send_job(grid=self.circuit, instruction=instruction)

        else:
            # declare the allowed file types
            files_types = ("VeraGrid zip (*.veragrid);;"                        
                           "VeraGrid HDF5 (*.gch5);;"
                           "Excel (*.xlsx);;"
                           "CGMES (*.zip);;"
                           "CIM (*.xml);;"
                           "Electrical Json V3 (*.ejson3);;"
                           "Raw (*.raw);;"
                           "Rawx (*.rawx);;"
                           "DIgSILENT Power Factory DGS (*.dgs);;"
                           "Sqlite (*.sqlite);;")

            if NEWTON_PA_AVAILABLE:
                files_types += "Newton (*.newton);;"

            if PGM_AVAILABLE:
                files_types += "PGM Json (*.pgm);;"

            # call dialog to select the file
            if self.project_directory is None:
                self.project_directory = ''

            # gather comments
            self.circuit.comments = self.ui.comments_textEdit.toPlainText()

            if self.file_name == '':
                # if the global file_name is empty, ask where to save
                fname = os.path.join(self.project_directory, self.ui.grid_name_line_edit.text())

                filename, type_selected = QtWidgets.QFileDialog.getSaveFileName(self, 'Save file', fname, files_types)

                if filename != '':

                    # if the user did not enter the extension, add it automatically
                    name, file_extension = os.path.splitext(filename)

                    extension = dict()
                    extension['Excel (*.xlsx)'] = '.xlsx'
                    extension['CIM (*.xml)'] = '.xml'
                    extension['CGMES (*.zip)'] = '.zip'
                    extension['Electrical Json V2 (*.ejson2)'] = '.ejson2'
                    extension['Electrical Json V3 (*.ejson3)'] = '.ejson3'
                    extension['VeraGrid zip (*.veragrid)'] = '.veragrid'
                    extension['Raw (*.raw)'] = '.raw'
                    extension['Rawx (*.rawx)'] = '.rawx'
                    extension['VeraGrid HDF5 (*.gch5)'] = '.gch5'
                    extension['Sqlite (*.sqlite)'] = '.sqlite'
                    extension['Newton (*.newton)'] = '.newton'
                    extension['PGM Json (*.pgm)'] = '.pgm'
                    extension['DIgSILENT Power Factory DGS (*.dgs)'] = '.dgs'

                    if file_extension == '':
                        filename = name + extension[type_selected]

                    # we were able to compose the file correctly, now save it
                    self.file_name = filename
                    self.save_file_now(self.file_name, type_selected=type_selected)
            else:
                # save directly
                self.save_file_now(self.file_name)

    def get_file_save_options(self) -> filedrv.FileSavingOptions:
        """
        Compose the file saving options
        :return: FileSavingOptions
        """

        # get save data
        sessions_data = self.session.get_save_data() if self.ui.saveResultsCheckBox.isChecked() else list()

        # get json files to store
        json_files = {"gui_config": self.get_gui_config_data()}

        cgmes_version = self.cgmes_version_dict[self.ui.cgmes_version_comboBox.currentText()]
        cgmes_profiles_txt = gf.get_checked_values(mdl=self.ui.cgmes_profiles_listView.model())
        cgmes_profiles = [self.cgmes_profiles_dict[e] for e in cgmes_profiles_txt]
        cgmes_one_file_per_profile = self.ui.cgmes_single_profile_per_file_checkBox.isChecked()
        cgmes_map_areas_like_raw = self.ui.cgmes_map_regions_like_raw_checkBox.isChecked()

        raw_version = self.ui.raw_export_version_comboBox.currentText()

        options = filedrv.FileSavingOptions(cgmes_boundary_set=self.current_boundary_set,
                                            sessions_data=sessions_data,
                                            dictionary_of_json_files=json_files,
                                            cgmes_version=cgmes_version,
                                            cgmes_profiles=cgmes_profiles,
                                            cgmes_one_file_per_profile=cgmes_one_file_per_profile,
                                            cgmes_map_areas_like_raw=cgmes_map_areas_like_raw,
                                            raw_version=raw_version)

        return options

    def get_file_open_options(self) -> filedrv.FileOpenOptions:
        """
        Compose the file open options
        :return: FileOpenOptions
        """

        cgmes_map_areas_like_raw = self.ui.cgmes_map_regions_like_raw_checkBox.isChecked()
        try_to_map_dc_to_hvdc_line = self.ui.cgmes_dc_as_hvdclines_checkBox.isChecked()

        options = filedrv.FileOpenOptions(cgmes_map_areas_like_raw=cgmes_map_areas_like_raw,
                                          try_to_map_dc_to_hvdc_line=try_to_map_dc_to_hvdc_line)

        return options

    def save_file_now(self, filename: str, type_selected: str = "", grid: Union[MultiCircuit, None] = None):
        """
        Save the file right now, without questions
        :param filename: filename to save to
        :param type_selected: File type description as it appears
                              in the file saving dialogue i.e. VeraGrid zip (*.veragrid)
        :param grid: MultiCircuit or None, if None, self.circuit is taken
        """

        if ('file_save' not in self.stuff_running_now) and ('file_open' not in self.stuff_running_now):
            # lock the ui
            self.LOCK()

            # check not to kill threads avoiding segmentation faults
            if self.save_file_thread_object is not None:
                if self.save_file_thread_object.isRunning():
                    ok = yes_no_question("There is a saving procedure running.\nCancel and retry?")
                    if ok:
                        self.save_file_thread_object.quit()

            options = self.get_file_save_options()
            options.type_selected = type_selected

            self.save_file_thread_object = filedrv.FileSaveThread(circuit=self.circuit if grid is None else grid,
                                                                  file_name=filename,
                                                                  options=options)

            # make connections
            self.save_file_thread_object.progress_signal.connect(self.ui.progressBar.setValue)
            self.save_file_thread_object.progress_text.connect(self.ui.progress_label.setText)
            self.save_file_thread_object.done_signal.connect(self.UNLOCK)
            self.save_file_thread_object.done_signal.connect(self.post_file_save)

            # thread start
            self.save_file_thread_object.start()

            # register as the latest file driver
            self.last_file_driver = self.save_file_thread_object

            self.stuff_running_now.append(SimulationTypes.FileSave)

        else:
            warning_msg('There is a file being processed..')

    def post_file_save(self):
        """
        Actions after the threaded file save
        """
        if self.save_file_thread_object.logger is not None:
            if len(self.save_file_thread_object.logger) > 0:
                dlg = LogsDialogue('Save file logger', self.save_file_thread_object.logger)
                dlg.exec()

        self.stuff_running_now.remove(SimulationTypes.FileSave)

        self.ui.model_version_label.setText('Model v. ' + str(self.circuit.model_version))
        self.ui.grid_idtag_label.setText('idtag. ' + str(self.circuit.idtag))

        # get the session tree structure
        session_data_dict = self.save_file_thread_object.get_session_tree()
        mdl = gf.get_tree_model(session_data_dict, 'Sessions')
        self.ui.diskSessionsTreeView.setModel(mdl)

        # call the garbage collector to free memory
        self.collect_memory()

    def grid_generator(self):
        """
        Open the grid generator window
        """
        self.grid_generator_dialogue = GridGeneratorGUI(parent=self)
        self.grid_generator_dialogue.resize(int(1.61 * 600.0), 550)  # golden ratio
        self.grid_generator_dialogue.exec()

        if self.grid_generator_dialogue.applied:

            if self.circuit.valid_for_simulation() > 0:
                reply = QtWidgets.QMessageBox.question(self, 'Message',
                                                       'Are you sure that you want to delete_with_dialogue '
                                                       'the current grid and replace it?',
                                                       QtWidgets.QMessageBox.StandardButton.Yes,
                                                       QtWidgets.QMessageBox.StandardButton.No)

                if reply == QtWidgets.QMessageBox.StandardButton.No:
                    return

            self.circuit = self.grid_generator_dialogue.circuit

            # create schematic
            self.redraw_current_diagram()

            # set circuit name
            diagram = self.get_selected_diagram_widget()
            if diagram is not None:
                if isinstance(diagram, SchematicWidget):
                    diagram.name.setText(f"Random grid {self.circuit.get_bus_number()} buses")

            # set base magnitudes
            self.ui.sbase_doubleSpinBox.setValue(self.circuit.Sbase)
            self.ui.fbase_doubleSpinBox.setValue(self.circuit.fBase)
            self.ui.model_version_label.setText(f"Model v. {self.circuit.model_version}")
            self.ui.grid_idtag_label.setText('idtag. ' + str(self.circuit.idtag))

            # set circuit comments
            self.ui.comments_textEdit.setText("Grid generated randomly using the RPGM algorithm.")

            # update the drop down menus that display dates
            self.update_date_dependent_combos()
            self.update_from_to_list_views()

            # clear the results
            self.clear_results()

    def import_bus_coordinates(self):
        """

        :return:
        """
        self.coordinates_window = CoordinatesInputGUI(self, self.circuit.get_buses())
        self.coordinates_window.exec()
        self.set_xy_from_lat_lon()

    def export_object_profiles(self):
        """
        Export object profiles
        """
        if self.circuit.time_profile is not None:

            # declare the allowed file types
            files_types = "Excel file (*.xlsx)"
            # call dialog to select the file
            if self.project_directory is None:
                self.project_directory = ''

            fname = os.path.join(self.project_directory, 'profiles of ' + self.ui.grid_name_line_edit.text())

            filename, type_selected = QtWidgets.QFileDialog.getSaveFileName(self, 'Save file', fname, files_types)

            if filename != "":
                if not filename.endswith('.xlsx'):
                    filename += '.xlsx'
                # TODO: correct this function
                self.circuit.export_profiles(file_name=filename)
        else:
            warning_msg('There are no profiles!', 'Export object profiles')

    def export_all(self):
        """
        Export all the results
        :return:
        """

        available_results = self.get_available_drivers()

        if len(available_results) > 0:

            files_types = "Zip file (*.zip)"
            fname = os.path.join(self.project_directory, 'Results of ' + self.ui.grid_name_line_edit.text())

            filename, type_selected = QtWidgets.QFileDialog.getSaveFileName(self, 'Save file', fname, files_types)

            if filename != "":
                self.LOCK()

                self.stuff_running_now.append(SimulationTypes.ExportAll)
                self.export_all_thread_object = exprtdrv.ExportAllThread(circuit=self.circuit,
                                                                         drivers_list=available_results,
                                                                         file_name=filename)

                self.export_all_thread_object.progress_signal.connect(self.ui.progressBar.setValue)
                self.export_all_thread_object.progress_text.connect(self.ui.progress_label.setText)
                self.export_all_thread_object.done_signal.connect(self.post_export_all)
                self.export_all_thread_object.start()
        else:
            warning_msg('There are no result available :/')

    def post_export_all(self):
        """
        Actions post export all
        """
        self.stuff_running_now.remove(SimulationTypes.ExportAll)

        if self.export_all_thread_object is not None:
            if self.export_all_thread_object.logger.has_logs():
                dlg = LogsDialogue('Export all', self.export_all_thread_object.logger)
                dlg.exec()

        if len(self.stuff_running_now) == 0:
            self.UNLOCK()

    def export_simulation_data(self):
        """
        Export the calculation objects to file
        """

        # declare the allowed file types
        files_types = "Excel file (*.xlsx)"

        # call dialog to select the file
        if self.project_directory is None:
            self.project_directory = ''

        fname = os.path.join(self.project_directory, self.ui.grid_name_line_edit.text())

        filename, type_selected = QtWidgets.QFileDialog.getSaveFileName(self, 'Save file', fname, files_types)

        if filename != "":
            if not filename.endswith('.xlsx'):
                filename += '.xlsx'

            if self.compiled_arrays is None:
                self.recompile_circuits_for_display()

            self.compiled_arrays.export_to_excel(file_name=filename)

            self.show_info_toast("Done!")

    def load_results_driver(self):
        """
        Load a driver from disk
        """
        idx = self.ui.diskSessionsTreeView.selectedIndexes()
        if len(idx) > 0:
            tree_mdl = self.ui.diskSessionsTreeView.model()
            item = tree_mdl.itemFromIndex(idx[0])
            path = gf.get_tree_item_path(item)

            if len(path) > 1:
                session_name = path[0]
                study_name = path[1]
                if self.last_file_driver is not None:
                    data_dict = self.last_file_driver.load_session_objects(session_name=session_name,
                                                                           study_name=study_name)

                    logger = self.session.register_driver_from_disk_data(grid=self.circuit,
                                                                         study_name=study_name,
                                                                         data_dict=data_dict)

                    if logger.has_logs():
                        dlg = LogsDialogue(name="Results parsing", logger=logger, expand_all=True)
                        dlg.exec()

                    self.update_available_results()
                else:
                    error_msg('No file driver declared :/')
            else:
                info_msg('Select a driver inside a session', 'Driver load from disk')

    def import_contingencies(self):
        """
        Open file to import contingencies file
        """

        files_types = "Formats (*.json)"

        # call dialog to select the file

        filenames, type_selected = QtWidgets.QFileDialog.getOpenFileNames(parent=self,
                                                                          caption='Open file',
                                                                          dir=self.project_directory,
                                                                          filter=files_types)

        if len(filenames) == 1:
            contingencies = import_contingencies_from_json(filenames[0])
            logger = self.circuit.set_contingencies(contingencies=contingencies)

            if len(logger) > 0:
                dlg = LogsDialogue('Contingencies import', logger)
                dlg.exec()

    def export_contingencies(self):
        """
        Export contingencies
        :return:
        """
        if len(self.circuit.contingencies) > 0:

            # declare the allowed file types
            files_types = "JSON file (*.json)"

            # call dialog to select the file
            filename, type_selected = QtWidgets.QFileDialog.getSaveFileName(self, 'Save file', '', files_types)

            if not (filename.endswith('.json')):
                filename += ".json"

            if filename != "":
                # save file
                export_contingencies_json_file(circuit=self.circuit, file_path=filename)

    def add_default_catalogue(self) -> None:
        """
        Add default catalogue to circuit
        """

        self.circuit.transformer_types += get_transformer_catalogue()
        self.circuit.underground_cable_types += get_cables_catalogue()
        self.circuit.wire_types += get_wires_catalogue()
        self.circuit.sequence_line_types += get_sequence_lines_catalogue()
        self.circuit.rms_models += get_rms_model_catalogue()

    def load_custom_catalogue(self):
        """
        Load a catalogue file and add it to the current one
        """
        # this will be filled with: open dialogue tab only, then connect select_csv_file from there
        """
        Open select component window for uploading catalogue data
        """

        files_types = "Catalogue file (*.xlsx)"

        filename, type_selected = QtWidgets.QFileDialog.getOpenFileName(parent=self,
                                                                        caption="Load catalogue",
                                                                        dir=self.project_directory,
                                                                        filter=files_types)

        if len(filename) > 0:
            if os.path.exists(filename):

                data, logger = load_catalogue(fname=filename)

                if logger.has_logs():
                    dlg = LogsDialogue('Open catalogue logger', logger)
                    dlg.exec()

                self.circuit.add_catalogue(data)
        else:
            return None

    def save_custom_catalogue(self):
        """
        Save the current catalogue
        """

        # declare the allowed file types
        files_types = "Catalogue file (*.xlsx)"

        # call dialog to select the file
        filename, type_selected = QtWidgets.QFileDialog.getSaveFileName(self,
                                                                        'Save catalogue', '', files_types)

        if not (filename.endswith('.xlsx')):
            filename += ".xlsx"

        if filename != "":
            save_catalogue(fname=filename, grid=self.circuit)

    def set_circuit(self, grid: MultiCircuit, create_diagram: bool = True):
        """

        :param grid:
        :param create_diagram:
        :return:
        """
        self.remove_all_diagrams()

        self.circuit = grid

        if create_diagram:
            self.add_complete_bus_branch_diagram()

        self.update_date_dependent_combos()
        self.update_from_to_list_views()
        self.clear_results()
        self.collect_memory()
        self.setup_time_sliders()
        self.get_circuit_snapshot_datetime()
        self.change_theme_mode()

    def cgmes_version_change(self):
        """
        On GUI cgmes version change, display only the supported profiles
        """
        cgmes_version = self.cgmes_version_dict[self.ui.cgmes_version_comboBox.currentText()]

        self.cgmes_profiles_dict = {key: CgmesProfileType(key) for key, val in
                                    get_available_cgmes_profiles(cgmes_version=cgmes_version).items()}

        self.ui.cgmes_profiles_listView.setModel(gf.get_list_model(list(self.cgmes_profiles_dict.keys()),
                                                                   checks=True, check_value=True))


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
from typing import Sequence
import numpy as np
import pandas as pd
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.Simulations.PowerFlow.Formulations.pf_advanced_formulation import (
    PfAdvancedFormulation)
from VeraGridEngine.Simulations.PowerFlow.power_flow_options import PowerFlowOptions
from VeraGridEngine.Compilers.circuit_to_data import compile_numerical_circuit_at
from VeraGridEngine.enumerations import EngineType, BusMode, BranchImpedanceMode
from VeraGridEngine.basic_structures import Logger


def PackDf(data: np.ndarray,
           columns: Sequence,
           index: Sequence,
           enumerate_index: bool = True,
           enumerate_cols: bool = False) -> pd.DataFrame:
    """
    Function to wrap DataFrame creation with index re-enumeration for easy reading
    :param data:
    :param columns:
    :param index:
    :param enumerate_index:
    :param enumerate_cols:
    :return: DataFrame
    """
    return pd.DataFrame(data=data,
                        columns=[f"{i}:{val}" for i, val in enumerate(columns)] if enumerate_cols else columns,
                        index=[f"{i}:{val}" for i, val in enumerate(index)] if enumerate_index else index
                        )


class CompiledArraysModule:
    available_structures = {
        "Bus arrays": [
            'V', 'Va', 'Vm',
            'S', 'P', 'Q',
            'I',
            'Y',
            'Qmin',
            'Qmax',
        ],
        "Bus indices": [
            'Types',
            'bus_ctrl',
            'pq',
            'pqv',
            'p',
            'pv',
            'vd',
            # 'pqpv',
        ],
        "Branch arrays": [
            'r',
            'x',
            'g',
            'b',
            'tap_f',
            'tap_t',
            'Pf_set',
            'Qf_set',
            'Pt_set',
            'Qt_set',
        ],

        "System matrices": [
            'Ybus',
            'G',
            'B',
            'Yf',
            'Yt',
            'Bbus',
            'Bf',
            'Cf',
            'Ct',
            "B'",
            "B''",
            'Yshunt',
            'Yseries',
        ],
        "Power flow arrays": [
            'idx_dPf',
            'idx_dQf',
            'idx_dPt',
            'idx_dQt',
            'idx_dVa',
            'idx_dVm',
            'idx_dm',
            'idx_dtau',
            'x0',
            'f(x0)',
            'Jacobian',
        ]
    }

    def __init__(self, grid: MultiCircuit,
                 t_idx: int | None,
                 engine: EngineType,
                 apply_temperature=False,
                 branch_tolerance_mode=BranchImpedanceMode.Specified,
                 use_stored_guess=False,
                 control_taps_modules: bool = True,
                 control_taps_phase: bool = True,
                 control_remote_voltage: bool = True,
                 fill_gep: bool = False,
                 fill_three_phase: bool = False):
        """

        :param grid:
        :param t_idx:
        :param engine:
        :param apply_temperature:
        :param branch_tolerance_mode:
        :param use_stored_guess:
        :param control_taps_modules:
        :param control_taps_phase:
        :param control_remote_voltage:
        :param fill_gep:
        :param fill_three_phase:
        """
        self.grid = grid

        if engine == EngineType.VeraGrid:
            nc = compile_numerical_circuit_at(
                circuit=self.grid,
                t_idx=t_idx,
                apply_temperature=apply_temperature,
                branch_tolerance_mode=branch_tolerance_mode,
                use_stored_guess=use_stored_guess,
                control_taps_phase=control_taps_phase,
                control_taps_modules=control_taps_modules,
                control_remote_voltage=control_remote_voltage,
                fill_gep=fill_gep,
                fill_three_phase=fill_three_phase,
            )
            self.islands = nc.split_into_islands()

        elif engine == EngineType.Bentayga:
            import VeraGridEngine.Compilers.circuit_to_bentayga as ben
            self.calculation_inputs_to_display = ben.get_snapshots_from_bentayga(self.grid)

        elif engine == EngineType.NewtonPA:
            import VeraGridEngine.Compilers.circuit_to_newton_pa as ne
            self.calculation_inputs_to_display = ne.get_snapshots_from_newtonpa(self.grid)

        else:
            # fallback to VeraGrid
            nc = compile_numerical_circuit_at(self.grid)
            self.islands = nc.split_into_islands()

    def get_structure(self, idx: int, structure_type: str) -> pd.DataFrame:
        """
        Get a DataFrame with the input.
        :param: structure_type: String representing structure type
        :return: pandas DataFrame
        """
        island = self.islands[idx]

        Sbus = island.get_power_injections_pu()
        idx = island.get_simulation_indices(Sbus=Sbus)

        Qmax_bus, Qmin_bus = island.get_reactive_power_limits()

        formulation = PfAdvancedFormulation(V0=island.bus_data.Vbus,
                                            S0=Sbus,
                                            I0=island.get_current_injections_pu(),
                                            Y0=island.get_admittance_injections_pu(),
                                            Qmin=Qmin_bus,
                                            Qmax=Qmax_bus,
                                            nc=island,
                                            options=PowerFlowOptions(),
                                            logger=Logger())

        if structure_type == 'V':
            df = PackDf(
                data=island.bus_data.Vbus,
                columns=['Voltage (p.u.)'],
                index=island.bus_data.names,
            )

        elif structure_type == 'Va':
            df = PackDf(
                data=np.angle(island.bus_data.Vbus),
                columns=['Voltage angles (rad)'],
                index=island.bus_data.names,
            )
        elif structure_type == 'Vm':
            df = PackDf(
                data=np.abs(island.bus_data.Vbus),
                columns=['Voltage modules (p.u.)'],
                index=island.bus_data.names,
            )
        elif structure_type == 'S':
            df = PackDf(
                data=Sbus,
                columns=['Power (p.u.)'],
                index=island.bus_data.names,
            )

        elif structure_type == 'P':
            df = PackDf(
                data=Sbus.real,
                columns=['Power (p.u.)'],
                index=island.bus_data.names,
            )

        elif structure_type == 'Q':
            df = PackDf(
                data=Sbus.imag,
                columns=['Power (p.u.)'],
                index=island.bus_data.names,
            )

        elif structure_type == 'I':
            df = PackDf(
                data=island.get_current_injections_pu(),
                columns=['Current (p.u.)'],
                index=island.bus_data.names,
            )

        elif structure_type == 'Y':
            df = PackDf(
                data=island.get_admittance_injections_pu(),
                columns=['Admittance (p.u.)'],
                index=island.bus_data.names,
            )

        elif structure_type == 'Ybus':
            adm = island.get_admittance_matrices()
            df = PackDf(
                data=adm.Ybus.toarray(),
                columns=island.bus_data.names,
                index=island.bus_data.names,
            )

        elif structure_type == 'G':
            adm = island.get_admittance_matrices()
            df = PackDf(
                data=adm.Ybus.real.toarray(),
                columns=island.bus_data.names,
                index=island.bus_data.names,
            )

        elif structure_type == 'B':
            adm = island.get_admittance_matrices()
            df = PackDf(
                data=adm.Ybus.imag.toarray(),
                columns=island.bus_data.names,
                index=island.bus_data.names,
            )

        elif structure_type == 'Yf':
            adm = island.get_admittance_matrices()
            df = PackDf(
                data=adm.Yf.toarray(),
                columns=island.bus_data.names,
                index=island.passive_branch_data.names,
            )

        elif structure_type == 'Yt':
            adm = island.get_admittance_matrices()
            df = PackDf(
                data=adm.Yt.toarray(),
                columns=island.bus_data.names,
                index=island.passive_branch_data.names,
            )

        elif structure_type == 'Bbus':
            adm = island.get_linear_admittance_matrices(idx)
            df = PackDf(
                data=adm.Bbus.toarray(),
                columns=island.bus_data.names,
                index=island.bus_data.names,
            )

        elif structure_type == 'Bf':
            adm = island.get_linear_admittance_matrices(idx)
            df = PackDf(
                data=adm.Bf.toarray(),
                columns=island.bus_data.names,
                index=island.passive_branch_data.names,
            )

        elif structure_type == 'Cf':
            df = PackDf(
                data=island.passive_branch_data.Cf.toarray(),
                columns=island.bus_data.names,
                index=island.passive_branch_data.names,
            )

        elif structure_type == 'Ct':
            df = PackDf(
                data=island.passive_branch_data.Ct.toarray(),
                columns=island.bus_data.names,
                index=island.passive_branch_data.names,
            )

        elif structure_type == 'Yshunt':
            df = PackDf(
                data=island.get_Yshunt_bus_pu(),
                columns=['Shunt admittance (p.u.)'],
                index=island.bus_data.names,
            )

        elif structure_type == 'Yseries':
            adms = island.get_series_admittance_matrices()
            df = PackDf(
                data=adms.Yseries.toarray(),
                columns=island.bus_data.names,
                index=island.bus_data.names,
            )

        elif structure_type == "B'":
            adm = island.get_fast_decoupled_amittances()

            if adm.B1.shape[0] == len(idx.pq):
                data = adm.B1.toarray()
                names = island.bus_data.names[idx.pq]
            else:
                data = adm.B1[np.ix_(idx.pq, idx.pq)].toarray()
                names = island.bus_data.names[idx.pq]

            df = PackDf(
                data=data,
                columns=names,
                index=names,
            )

        elif structure_type == "B''":
            adm = island.get_fast_decoupled_amittances()

            if adm.B2.shape[0] == len(idx.pq):
                data = adm.B2.toarray()
                names = island.bus_data.names[idx.pq]
            else:
                data = adm.B2[np.ix_(idx.pq, idx.pq)].toarray()
                names = island.bus_data.names[idx.pq]

            df = PackDf(
                data=data,
                columns=names,
                index=names,
            )

        elif structure_type == 'Types':
            data = island.bus_data.bus_types
            df = PackDf(
                data=data,
                columns=['Bus types'],
                index=island.bus_data.names,
            )

        elif structure_type == 'x0':
            df = PackDf(
                data=formulation.var2x(),
                columns=['x0'],
                index=formulation.get_x_names(),
            )

        elif structure_type == 'f(x0)':
            df = PackDf(
                data=formulation.fx(),
                columns=['f(x0)'],
                index=formulation.get_fx_names(),
            )

        elif structure_type == 'Jacobian':
            df = formulation.get_jacobian_df(autodiff=False)

        elif structure_type == 'Qmin':
            df = PackDf(
                data=Qmin_bus,
                columns=['Qmin'],
                index=island.bus_data.names,
            )

        elif structure_type == 'Qmax':
            df = PackDf(
                data=Qmax_bus,
                columns=['Qmax'],
                index=island.bus_data.names,
            )

        elif structure_type == 'bus_ctrl':
            data1 = [BusMode.as_str(val) for val in island.bus_data.bus_types]

            df = PackDf(
                data=np.array(data1),
                columns=['bus_ctrl'],
                index=island.bus_data.names,
            )

        elif structure_type == 'branch_ctrl':

            data1 = [val.value if val != 0 else "-" for val in island.active_branch_data.tap_module_control_mode]
            data2 = [val.value if val != 0 else "-" for val in island.active_branch_data.tap_phase_control_mode]

            df = PackDf(
                data=np.c_[
                    island.passive_branch_data.F,
                    island.passive_branch_data.T,
                    island.active_branch_data.tap_controlled_buses,
                    data1,
                    data2
                ],
                columns=['bus F', 'bus T', 'V ctrl bus', 'm control', 'tau control'],
                index=[f"{k}) {name}" for k, name in enumerate(island.passive_branch_data.names)],
            )

        elif structure_type == 'pq':
            df = PackDf(
                data=idx.pq.astype(int).astype(str),
                columns=['pq'],
                index=island.bus_data.names[idx.pq],
            )

        elif structure_type == 'pv':
            df = PackDf(
                data=idx.pv.astype(int).astype(str),
                columns=['pv'],
                index=island.bus_data.names[idx.pv],
            )

        elif structure_type == 'pqv':
            df = PackDf(
                data=idx.pqv.astype(int).astype(str),
                columns=['pqv'],
                index=island.bus_data.names[idx.pqv],
            )

        elif structure_type == 'p':
            df = PackDf(
                data=idx.p.astype(int).astype(str),
                columns=['p'],
                index=island.bus_data.names[idx.p],
            )

        elif structure_type == 'vd':
            df = PackDf(
                data=idx.vd.astype(int).astype(str),
                columns=['vd'],
                index=island.bus_data.names[idx.vd],
            )

        elif structure_type == 'r':
            df = PackDf(
                data=island.passive_branch_data.R,
                columns=['Resistance (p.u.)'],
                index=island.passive_branch_data.names,
            )

        elif structure_type == 'x':
            df = PackDf(
                data=island.passive_branch_data.X,
                columns=['Reactance (p.u.)'],
                index=island.passive_branch_data.names,
            )

        elif structure_type == 'g':
            df = PackDf(
                data=island.passive_branch_data.G,
                columns=['Conductance (p.u.)'],
                index=island.passive_branch_data.names,
            )

        elif structure_type == 'b':
            df = PackDf(
                data=island.passive_branch_data.B,
                columns=['Susceptance (p.u.)'],
                index=island.passive_branch_data.names,
            )
        elif structure_type == 'tap_f':
            df = PackDf(
                data=island.passive_branch_data.virtual_tap_f,
                columns=['Virtual tap from (p.u.)'],
                index=island.passive_branch_data.names,
            )

        elif structure_type == 'tap_t':
            df = PackDf(
                data=island.passive_branch_data.virtual_tap_t,
                columns=['Virtual tap to (p.u.)'],
                index=island.passive_branch_data.names,
            )


        elif structure_type == 'idx_dPf':
            df = PackDf(
                data=formulation.idx_dPf.astype(int).astype(str),
                columns=['idx_dPf'],
                index=island.passive_branch_data.names[formulation.idx_dPf],
            )

        elif structure_type == 'idx_dQf':
            df = PackDf(
                data=formulation.idx_dQf.astype(int).astype(str),
                columns=['idx_dQf'],
                index=island.passive_branch_data.names[formulation.idx_dQf],
            )

        elif structure_type == 'idx_dPt':
            df = PackDf(
                data=formulation.idx_dPt.astype(int).astype(str),
                columns=['idx_dPt'],
                index=island.passive_branch_data.names[formulation.idx_dPt],
            )

        elif structure_type == 'idx_dQt':
            df = PackDf(
                data=formulation.idx_dQt.astype(int).astype(str),
                columns=['idx_dQt'],
                index=island.passive_branch_data.names[formulation.idx_dQt],
            )

        elif structure_type == 'idx_dVa':
            df = PackDf(
                data=formulation.idx_dVa.astype(int).astype(str),
                columns=['idx_dVa'],
                index=island.bus_data.names[formulation.idx_dVa],
            )

        elif structure_type == 'idx_dVm':
            df = PackDf(
                data=formulation.idx_dVm.astype(int).astype(str),
                columns=['idx_dVm'],
                index=island.bus_data.names[formulation.idx_dVm],
            )

        elif structure_type == 'idx_dm':
            df = PackDf(
                data=formulation.idx_dm.astype(int).astype(str),
                columns=['idx_dm'],
                index=island.passive_branch_data.names[formulation.idx_dm],
            )

        elif structure_type == 'idx_dtau':
            df = PackDf(
                data=formulation.idx_dtau.astype(int).astype(str),
                columns=['idx_dtau'],
                index=island.passive_branch_data.names[formulation.idx_dtau],
            )

        elif structure_type == 'Pf_set':
            df = PackDf(
                data=island.active_branch_data.Pset[formulation.idx_dPf],
                columns=['Pf_set'],
                index=island.passive_branch_data.names[formulation.idx_dPf],
            )

        elif structure_type == 'Pt_set':
            df = PackDf(
                data=island.active_branch_data.Pset[formulation.idx_dPt],
                columns=['Pt_set'],
                index=island.passive_branch_data.names[formulation.idx_dPt],
            )

        elif structure_type == 'Qf_set':
            df = PackDf(
                data=island.active_branch_data.Qset[formulation.idx_dQf],
                columns=['Qf_set'],
                index=island.passive_branch_data.names[formulation.idx_dQf],
            )

        elif structure_type == 'Qt_set':
            df = PackDf(
                data=island.active_branch_data.Qset[formulation.idx_dQt],
                columns=['Qt_set'],
                index=island.passive_branch_data.names[formulation.idx_dQt],
            )

        else:
            raise Exception('PF input: structure type not found' + str(structure_type))

        return df

    def export_to_excel(self, file_name: str):
        """
        Export all to Excel
        :param file_name: file name
        :return: 
        """
        if not file_name.endswith('.xlsx'):
            file_name += '.xlsx'

        with pd.ExcelWriter(file_name) as writer:  # pylint: disable=abstract-class-instantiated

            for c, calc_input in enumerate(self.islands):

                for category, elms_in_category in self.available_structures.items():
                    for elm_type in elms_in_category:
                        name = f"{category}_{elm_type}@{c}"
                        df = self.get_structure(idx=c, structure_type=elm_type).astype(str)

                        if isinstance(df, pd.DataFrame):
                            df.to_excel(excel_writer=writer,
                                        sheet_name=name[:31])  # excel supports 31 chars per sheet name




# MIT License
#
# Copyright (c) 2018 Ross Wilson
# Copyright (c) 2024, Santiago Peñate Vera
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
A tile source that serves BlueMarble tiles from the internet.
"""

import math
from typing import Tuple
from VeraGrid.Gui.Diagrams.MapWidget.Tiles.tiles import Tiles


class BlueMarbleTiles(Tiles):
    """
    An object to source internet tiles for pySlip.
    """

    def __init__(self,
                 tiles_dir='blue_marble_tiles',
                 http_proxy=None,
                 name: str = 'Blue Marble'):
        """
        Override the base class for these tiles.

        Basically, just fill in the BaseTiles class with values from above
        and provide the Geo2Tile() and Tile2Geo() methods.
        :param tiles_dir:
        :param http_proxy:
        """

        super().__init__(tile_set_name=name,
                         tile_set_short_name='BM Tiles',
                         tile_set_version='1.0',
                         levels=list(range(10)),
                         tile_width=256,
                         tile_height=256,
                         tiles_dir=tiles_dir,
                         max_lru=10000,
                         servers=['https://s3.amazonaws.com', ],
                         url_path='/com.modestmaps.bluemarble/{Z}-r{Y}-c{X}.jpg',
                         max_server_requests=2,
                         http_proxy=http_proxy,
                         attribution="© NASA Blue Marble, © OpenStreetMap contributors")

    def copy(self) -> "BlueMarbleTiles":
        """
        Copy of this object
        :return:
        """

        cpy = BlueMarbleTiles(tiles_dir=self.tiles_dir,
                              http_proxy=self.http_proxy,
                              name=self.tile_set_name)

        return cpy

    def Geo2Tile(self, longitude: float, latitude: float) -> Tuple[float, float]:
        """Convert geo to tile fractional coordinates for level in use.

        geo  tuple of geo coordinates (xgeo, ygeo)

        Note that we assume the point *is* on the map!

        Code taken from [https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames]
        """

        lat_rad = math.radians(latitude)
        n = 2.0 ** self.level
        xtile = (longitude + 180.0) / 360.0 * n
        ytile = ((1.0 - math.log(math.tan(lat_rad) + (1.0 / math.cos(lat_rad))) / math.pi) / 2.0) * n

        return xtile, ytile

    def Tile2Geo(self, x_tile: float, y_tile: float) -> Tuple[float, float]:
        """Convert tile fractional coordinates to geo for level in use.

        tile  a tupl;e (xtile,ytile) of tile fractional coordinates

        Note that we assume the point *is* on the map!

        Code taken from [https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames]
        """
        n = 2.0 ** self.level
        xgeo = x_tile / n * 360.0 - 180.0
        yrad = math.atan(math.sinh(math.pi * (1 - 2 * y_tile / n)))
        ygeo = math.degrees(yrad)

        return xgeo, ygeo



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'rms_model_editor.ui'
##
## Created by: Qt User Interface Compiler version 6.9.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QFrame, QHBoxLayout,
    QLabel, QMainWindow, QPushButton, QSizePolicy,
    QSpacerItem, QVBoxLayout, QWidget)
from VeraGrid.Gui.Icons.icons_rc import *

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(966, 538)
        self.actionCheckModel = QAction(MainWindow)
        self.actionCheckModel.setObjectName(u"actionCheckModel")
        self.actionCheckModel.setMenuRole(QAction.MenuRole.NoRole)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.frame = QFrame(self.centralwidget)
        self.frame.setObjectName(u"frame")
        self.frame.setMaximumSize(QSize(16777215, 40))
        self.frame.setFrameShape(QFrame.Shape.NoFrame)
        self.frame.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout = QHBoxLayout(self.frame)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.model_selector_comboBox = QComboBox(self.frame)
        self.model_selector_comboBox.addItem("")
        self.model_selector_comboBox.addItem("")
        self.model_selector_comboBox.addItem("")
        self.model_selector_comboBox.setObjectName(u"model_selector_comboBox")

        self.horizontalLayout.addWidget(self.model_selector_comboBox)

        self.horizontalSpacer = QSpacerItem(856, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.currently_editing_block_label = QLabel(self.frame)
        self.currently_editing_block_label.setObjectName(u"currently_editing_block_label")

        self.horizontalLayout.addWidget(self.currently_editing_block_label)

        self.horizontalSpacer_2 = QSpacerItem(48, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer_2)

        self.currently_editing_object_label = QLabel(self.frame)
        self.currently_editing_object_label.setObjectName(u"currently_editing_object_label")

        self.horizontalLayout.addWidget(self.currently_editing_object_label)


        self.verticalLayout.addWidget(self.frame)

        self.mainLayout = QVBoxLayout()
        self.mainLayout.setObjectName(u"mainLayout")

        self.verticalLayout.addLayout(self.mainLayout)

        self.frame_2 = QFrame(self.centralwidget)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setMaximumSize(QSize(16777215, 40))
        self.frame_2.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_2.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout_2 = QHBoxLayout(self.frame_2)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.InitGuessButton = QPushButton(self.frame_2)
        self.InitGuessButton.setObjectName(u"InitGuessButton")

        self.horizontalLayout_2.addWidget(self.InitGuessButton)

        self.horizontalSpacer_3 = QSpacerItem(859, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_3)

        self.doItButton = QPushButton(self.frame_2)
        self.doItButton.setObjectName(u"doItButton")

        self.horizontalLayout_2.addWidget(self.doItButton)


        self.verticalLayout.addWidget(self.frame_2)

        MainWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"MainWindow", None))
        self.actionCheckModel.setText(QCoreApplication.translate("MainWindow", u"CheckModel", None))
        self.model_selector_comboBox.setItemText(0, QCoreApplication.translate("MainWindow", u"Template model", None))
        self.model_selector_comboBox.setItemText(1, QCoreApplication.translate("MainWindow", u"Block editor model", None))
        self.model_selector_comboBox.setItemText(2, QCoreApplication.translate("MainWindow", u"Equations editor model", None))

        self.model_selector_comboBox.setPlaceholderText(QCoreApplication.translate("MainWindow", u"Select model to apply", None))
        self.currently_editing_block_label.setText(QCoreApplication.translate("MainWindow", u"Submodel:", None))
        self.currently_editing_object_label.setText(QCoreApplication.translate("MainWindow", u"Device:", None))
        self.InitGuessButton.setText(QCoreApplication.translate("MainWindow", u"Add Initial Values", None))
        self.doItButton.setText(QCoreApplication.translate("MainWindow", u"Do it!", None))
    # retranslateUi




# MIT License
#
# Copyright (c) 2021-2022 Yunosuke Ohsugi
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import VeraGrid.ThirdParty.darkdetect as darkdetect
from VeraGrid.ThirdParty.qdarktheme import _os_appearance
from VeraGrid.ThirdParty.qdarktheme.qtpy.QtCore import QCoreApplication, QEvent, QObject, QThread, Signal


class OSThemeSwitchListener(QThread):
    """Listener to detect to change OS's theme."""

    sig_run = Signal(bool)
    _sig_listen_os_theme = Signal()

    def __init__(self, callable) -> None:
        """Initialize listener."""
        super().__init__()
        self.setProperty("is_running", True)
        self._theme = darkdetect.theme()
        self._accent = _os_appearance.accent()
        self.sig_run.connect(lambda state: self.setProperty("is_running", state))
        self._sig_listen_os_theme.connect(callable)

    def eventFilter(self, q_object: QObject, event: QEvent) -> bool:  # noqa: N802
        """Override QObject.eventFilter.

        This override is for Mac.
        Qt can listen to change OS's theme on only mac and changed QPalette automatically.
        So enable to listen to change OS's theme via palette change event.
        """
        if (
            self.property("is_running")
            and q_object == QCoreApplication.instance()
            and event.type() == QEvent.Type.ApplicationPaletteChange
        ):
            accent = _os_appearance.accent()
            theme = darkdetect.theme()
            if self._theme != theme or self._accent != accent:
                self._theme = theme
                self._accent = accent
                self._sig_listen_os_theme.emit()
                return True
        return super().eventFilter(q_object, event)

    def run(self) -> None:
        """Override QThread.run.

        This override is for except Mac.
        Qt cannot listen to change OS's theme on except Mac.
        Use ``darkdetect.listener`` to detect to change OS's theme.
        """
        darkdetect.listener(
            lambda theme: self.property("is_running") and self._sig_listen_os_theme.emit()
        )

    def kill(self) -> None:
        """Kill thread."""
        self.terminate()
        self.deleteLater()



# MIT License
#
# Copyright (c) 2021-2022 Yunosuke Ohsugi
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Default color values."""

THEME_COLOR_VALUES = {
    "dark": '{"background": {"base": "#202124", "list": {}, "panel": {"darken": 0.3}, "popup": {"lighten": 0.3}, "table": {"darken": 0.5}, "textarea": {"darken": 0.13}, "title": {"darken": 0.3}}, "border": {"base": "#3f4042", "input": {"transparent": 0}}, "foreground": {"base": "#e4e7eb", "defaultButton.disabledBackground": {"transparent": 0.2}, "disabled": {"transparent": 0.4}, "disabledSelectionBackground": {"transparent": 0.2}, "icon": {"darken": 0.01}, "icon.unfocused": {"transparent": 0.6}, "input.placeholder": {"transparent": 0.6}, "progressBar.disabledBackground": {"transparent": 0.2}, "slider.disabledBackground": {"transparent": 0.2}, "sliderTrack.inactiveBackground": {"transparent": 0.1}}, "input.background": "#3f4042", "inputButton.hoverBackground": "#ffffff25", "linkVisited": "#c58af8", "list.alternateBackground": "#ffffff0c", "list.hoverBackground": "#ffffff13", "menubar.selectionBackground": "#ffffff25", "popupItem.checkbox.background": "#ffffff19", "popupItem.selectionBackground": "#ffffff22", "primary": {"base": "#8ab4f7", "button.activeBackground": {"darken": 0.14, "transparent": 0.23}, "button.hoverBackground": {"darken": 0.1, "transparent": 0.11}, "defaultButton.activeBackground": {"darken": 0.12}, "defaultButton.hoverBackground": {"darken": 0.06}, "list.inactiveSelectionBackground": {"lighten": 0.2, "transparent": 0.15}, "list.selectionBackground": {"darken": 0.2, "transparent": 0.4}, "progressBar.background": {"darken": 0.1}, "selection.background": {"darken": 0.2, "lighten": 0.1, "transparent": 0.4}, "sliderHandle.activeBackground": {"darken": 0.09}, "table.inactiveSelectionBackground": {"lighten": 0.2, "transparent": 0.18}, "table.selectionBackground": {"darken": 0.2, "transparent": 0.55}, "textarea.selectionBackground": {"darken": 0.2, "lighten": 0.1, "transparent": 0.4}}, "scrollbar.background": "#ffffff10", "scrollbarSlider.activeBackground": "#ffffff60", "scrollbarSlider.background": "#ffffff30", "scrollbarSlider.disabledBackground": "#ffffff15", "scrollbarSlider.hoverBackground": "#ffffff45", "statusBar.background": "#2a2b2e", "statusBarItem.activeBackground": "#ffffff34", "statusBarItem.hoverBackground": "#ffffff22", "tab.activeBackground": "#ffffff00", "tab.hoverBackground": "#ffffff18", "tabCloseButton.hoverBackground": "#ffffff25", "table.alternateBackground": "#ffffff15", "tableSectionHeader.background": "#3f4042", "textarea.inactiveSelectionBackground": "#ffffff20", "toolbar.activeBackground": "#ffffff34", "toolbar.background": "#333333", "toolbar.hoverBackground": "#ffffff22", "tree.inactiveIndentGuidesStroke": "#ffffff35", "tree.indentGuidesStroke": "#ffffff60", "treeSectionHeader.background": "#3f4042"}',  # noqa: E501
    "light": '{"background": {"base": "#f8f9fa", "list": {}, "panel": {"lighten": 0.5}, "popup": {"lighten": 0.2}, "table": {"lighten": 0.5}, "textarea": {"lighten": 0.25}, "title": {"darken": 0.04}}, "border": {"base": "#dadce0", "input": {}}, "foreground": {"base": "#4d5157", "defaultButton.disabledBackground": {"transparent": 0.25}, "disabled": {"transparent": 0.4}, "disabledSelectionBackground": {"transparent": 0.25}, "icon": {"darken": 0.05}, "icon.unfocused": {"transparent": 0.6}, "input.placeholder": {"transparent": 0.6}, "progressBar.disabledBackground": {"transparent": 0.25}, "slider.disabledBackground": {"transparent": 0.25}, "sliderTrack.inactiveBackground": {"transparent": 0.2}}, "input.background": "#f8f9fa", "inputButton.hoverBackground": "#00000018", "linkVisited": "#660098", "list.alternateBackground": "#00000009", "list.hoverBackground": "#00000013", "menubar.selectionBackground": "#00000020", "popupItem.checkbox.background": "#00000019", "popupItem.selectionBackground": "#00000022", "primary": {"base": "#1a73e8", "button.activeBackground": {"darken": 0.03, "transparent": 0.24}, "button.hoverBackground": {"darken": 0.03, "transparent": 0.1}, "defaultButton.activeBackground": {"lighten": 0.3}, "defaultButton.hoverBackground": {"lighten": 0.1}, "list.inactiveSelectionBackground": {"darken": 0.43, "transparent": 0.09}, "list.selectionBackground": {"lighten": 0.2, "transparent": 0.35}, "progressBar.background": {"lighten": 0.2}, "selection.background": {"lighten": 0.3, "transparent": 0.5}, "sliderHandle.activeBackground": {"lighten": 0.2}, "table.inactiveSelectionBackground": {"darken": 0.43, "transparent": 0.09}, "table.selectionBackground": {"lighten": 0.1, "transparent": 0.5}, "textarea.selectionBackground": {"lighten": 0.3, "transparent": 0.5}}, "scrollbar.background": "#00000010", "scrollbarSlider.activeBackground": "#00000060", "scrollbarSlider.background": "#00000040", "scrollbarSlider.disabledBackground": "#00000015", "scrollbarSlider.hoverBackground": "#00000050", "statusBar.background": "#dfe1e5", "statusBarItem.activeBackground": "#00000024", "statusBarItem.hoverBackground": "#00000015", "tab.activeBackground": "#00000000", "tab.hoverBackground": "#00000015", "tabCloseButton.hoverBackground": "#00000020", "table.alternateBackground": "#00000012", "tableSectionHeader.background": "#dadce0", "textarea.inactiveSelectionBackground": "#00000015", "toolbar.activeBackground": "#00000024", "toolbar.background": "#ebebeb", "toolbar.hoverBackground": "#00000015", "tree.inactiveIndentGuidesStroke": "#00000030", "tree.indentGuidesStroke": "#00000050", "treeSectionHeader.background": "#dadce0"}',  # noqa: E501
}
ACCENT_COLORS = {
    "dark": {
        "blue": "#8ab4f7",
        "graphite": "#898a8f",
        "green": "#4caf50",
        "orange": "#ff9800",
        "pink": "#c7457f",
        "purple": "#af52bf",
        "red": "#f6685e",
        "yellow": "#ffeb3b",
    },
    "light": {
        "blue": "#1a73e8",
        "graphite": "#898a8f",
        "green": "#4caf50",
        "orange": "#ff9800",
        "pink": "#c7457f",
        "purple": "#9c27b0",
        "red": "#f44336",
        "yellow": "#f4c65f",
    },
}



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations
from typing import TYPE_CHECKING

from PySide6 import QtWidgets
from PySide6.QtCore import QPointF
from PySide6.QtGui import QPolygonF
from VeraGrid.Gui.gui_functions import add_menu_entry
from VeraGrid.Gui.Diagrams.generic_graphics import Polygon
from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.injections_template_graphics import InjectionTemplateGraphicItem
from VeraGrid.Gui.RmsModelEditor.rms_model_editor_engine import RmsModelEditorGUI
from VeraGridEngine.Devices.Injections.load import Load
from VeraGridEngine.enumerations import DeviceType

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget


class LoadGraphicItem(InjectionTemplateGraphicItem):

    def __init__(self, parent, api_obj: Load, editor: SchematicWidget):
        """

        :param parent:
        :param api_obj:
        :param editor:
        """
        InjectionTemplateGraphicItem.__init__(self,
                                              parent=parent,
                                              api_obj=api_obj,
                                              editor=editor,
                                              device_type_name='load',
                                              w=20,
                                              h=20)

        # triangle
        self.set_glyph(glyph=Polygon(
            self,
            polygon=QPolygonF([QPointF(0, 0),
                               QPointF(self.w, 0),
                               QPointF(self.w / 2, self.h)]),
            update_nexus_fcn=self.update_nexus)
        )

    @property
    def api_object(self) -> Load:
        return self._api_object

    def contextMenuEvent(self, event: QtWidgets.QGraphicsSceneContextMenuEvent):
        """
        Display context menu
        @param event:
        @return:
        """
        if self.api_object is not None:
            menu = self.get_base_context_menu()
            menu.addSection("Load")

            add_menu_entry(menu=menu,
                           text="Rms Editor",
                           function_ptr=self.edit_rms,
                           icon_path=":/Icons/icons/edit.png")

            menu.exec(event.screenPos())
        else:
            self.editor.gui.show_error_toast("The graphic has no API object!")

    def edit_rms(self):
        """

        :return:
        """
        # load templates
        templates = self.editor.circuit.rms_models

        # select line templates
        templ_catalogue = dict()
        templ_list = []
        for templ in templates:
            if templ.tpe == DeviceType.LoadDevice:
                templ_list.append(templ.name)
                templ_catalogue[templ.name] = templ

        # prompt RmsModelEditorGUI
        rms_model_editor = RmsModelEditorGUI(api_object_model_host=self.api_object.rms_model, templates_list=templ_list,
                                             templates_catalogue=templ_catalogue, api_object_name=self.api_object.name,
                                             api_object=self.api_object, main_editor= True, parent=self.editor)
        rms_model_editor.show()
